import re

from boimmgpy.id_converters.reactions_id_converter_scraper import ReactionsIDConverterScraper
from boimmgpy.utilities import file_utilities
from boimmgpy.id_converters.id_converter import IDConverter
from boimmgpy.definitions import TOOL_CONFIG_PATH, ROOT_DIR, DATABASE_CONFIGS


class ReactionsIDConverter(IDConverter):

    def __init__(self):
        super().__init__()

        # scraper = ReactionsIDConverterScraper()
        # path = scraper()
        self.__database_config = file_utilities.read_conf_file(DATABASE_CONFIGS)

        path = ROOT_DIR + self.__database_config["path_model_seed_reactions_aliases"]
        self.__configs = file_utilities.read_conf_file(
            TOOL_CONFIG_PATH)
        self.__home_path__ = ROOT_DIR
        self.construct_reactions_converter(path)


    def convert_modelSeedId_into_other_dbID_reaction(self,modelSeedId,database_name):
        return self.modelSeedToDb.get(modelSeedId).get(database_name)

    def convert_dbID_into_modelSeedId_reaction(self,database_name,id):
        if id in self.dbToModelSeed.get(database_name).keys():
            return self.dbToModelSeed.get(database_name).get(id)
        else:
            return None

    def construct_reactions_converter(self,path):

        # try:
        #     info = urllib.request.urlopen("https://raw.githubusercontent.com/ModelSEED/ModelSEEDDatabase/dev/Biochemistry/Aliases/Unique_ModelSEED_Reaction_Aliases.txt")
        #     url = True
        # except:
        info = open(path, "r")
        extra_info = open(self.__home_path__ + self.__configs.get("reaction_converter_extra"), "r")
        url=False

        previousModelSeed = ""
        previousOtherDb = ""
        new_dic={}
        for line in info:

            if url:
                line=line.decode("utf-8")

            l = line.split("\t")


            modelSeedId = l[0]
            externalId = l[1]
            source = l[2].strip()

            if re.search("\AModelSEED", line) == None:

                if previousModelSeed != modelSeedId:
                    self.modelSeedToDb[previousModelSeed] = new_dic
                    new_dic = {}
                    new_dic[source]=[externalId]
                    previousModelSeed = modelSeedId
                    previousOtherDb = source

                elif previousOtherDb != source:
                    new_dic[source]=[externalId]
                    previousOtherDb=source

                else:
                    new_dic[source].append(externalId)

                if source in self.dbToModelSeed.keys():
                    if externalId in self.dbToModelSeed.get(source).keys():
                        self.dbToModelSeed[source][externalId].append(modelSeedId)
                    else:
                        self.dbToModelSeed[source][externalId] = [modelSeedId]

                else:
                    dic = {}
                    dic[externalId] = [modelSeedId]
                    self.dbToModelSeed[source] = dic
        info.close()
        for line in extra_info:

            if url:
                line = line.decode("utf-8")

            l = line.split("\t")

            modelSeedId = l[0]
            externalId = l[1]
            source = l[2].strip()

            if re.search("\AModelSEED", line) == None:

                if modelSeedId in self.modelSeedToDb.keys():
                    if source in self.modelSeedToDb.get(modelSeedId) and \
                        externalId not in self.modelSeedToDb[modelSeedId][source] and externalId:
                        self.modelSeedToDb[modelSeedId][source].append(externalId)
                    elif externalId:
                        self.modelSeedToDb[modelSeedId][source] = [externalId]

                if source in self.dbToModelSeed.keys():
                    if externalId in self.dbToModelSeed.get(source).keys():
                        self.dbToModelSeed[source][externalId].append(modelSeedId)
                    else:
                        self.dbToModelSeed[source][externalId] = [modelSeedId]

                else:
                    dic = {}
                    dic[externalId] = [modelSeedId]
                    self.dbToModelSeed[source] = dic

        extra_info.close()


from abc import ABC

from neo4j import GraphDatabase

from boimmgpy.database.interfaces import node
from boimmgpy.database.interfaces.boimmg_database_accessor import BOIMMGDatabaseAccessor


class ReactionsDBAccessor():

    def login(self):
        uri = "bolt://palsson.di.uminho.pt:6094"
        driver = GraphDatabase.driver(uri, auth=("neo4j", "Tuning999"), encrypted=False)
        self.tx = driver

    @property
    def tx(self):
        return self._tx

    @tx.setter
    def tx(self, tx):
        self._tx = tx

    def add_relationship(self, origin: int, target: int, type: str):
        self.login()
        with self.tx.session() as session:
            session.run("MATCH (c:Reaction), (p:Reaction)"
                        "WHERE ID(c) = $target and ID(p) = $origin "
                        "MERGE (c)<-[r:" + type + "]-(p) "
                                                  "ON CREATE SET r.timestamp = timestamp() ",
                        origin=origin, target=target, type=type)

    def get_predecessors_by_ont_id(self, ont_id: int) -> list:
        """Get predecessors using as parameter the database identifier"""
        pass

    def get_predecessors_by_ont_id_rel_type(self, ont_id: int, relationship_type : str) -> list:
        """Get predecessors using as parameter the database identifier and the relationship type"""

        raise NotImplementedError

    def get_node_by_ont_id(self, ont_id: int) -> node:
        """Get predecessors using as parameter the database identifier and the relationship type"""
        raise NotImplementedError


    def add_reaction(self,**kwargs):
        self.login()

        name = kwargs.get("name")

        onto_stoichiometry = kwargs.get("onto_stoichiometry")

        ms_stoich = kwargs.get("ms_stoich")

        direction = kwargs.get("direction")

        equation = self.generate_equation(onto_stoichiometry, ms_stoich, direction)

        ec_numbers = kwargs.get("ec_numbers") #list of ec numbers

        annotated = kwargs.get("annotated")

        deltaG = kwargs.get("deltaG")

        # aliases = kwargs.get("aliases")
        if "model_seed_id" in kwargs:
            model_seed_id = kwargs.get("model_seed_id")
        else:
            model_seed_id = None

        with self.tx.session() as session:

            if model_seed_id:
                result = session.run("merge (r:Reaction {equation: $equation, ec_numbers: $ec_numbers, "
                            "model_seed_id: $model_seed_id, annotated: $annotated}) "
                             "on create set r.name = $name, r.deltaG = $deltaG "
                            "return id(r) as r_id",
                            name = name, equation = equation, ec_numbers = ec_numbers,
                             annotated = annotated, deltaG = deltaG, model_seed_id = model_seed_id)
            else:
                result = session.run("merge (r:Reaction {equation: $equation, ec_numbers: $ec_numbers, "
                                     "annotated: $annotated}) "
                                     "on create set r.name = $name, r.deltaG = $deltaG "
                                     "return id(r) as r_id",
                                     name=name, equation=equation, ec_numbers=ec_numbers,
                                     annotated=annotated, deltaG=deltaG)

            result_reaction = result.single()
            reaction_id = result_reaction.get("r_id")

            for metabolite in onto_stoichiometry:
                coef = onto_stoichiometry[metabolite]

                if coef < 0:
                    role = "reactant"

                else:
                    role = "product"

                session.run("match (c:Compound), (r:Reaction) "
                            "where c.boimmg_id = $compound_id and id(r) = $reaction_id "
                            "merge (c)-[s:is_intervenient_in {coef: $coef, role : $role}]->(r) "
                            "on create set s.timestamp = timestamp()",
                            coef = coef, role = role, compound_id = metabolite,
                            reaction_id = reaction_id)

            for metabolite in ms_stoich:
                coef = ms_stoich[metabolite]

                if coef < 0:
                    role = "reactant"

                else:
                    role = "product"

                session.run("match (r:Reaction) "
                            "where id(r) = $reaction_id "
                            "merge (c:ModelSeedCompound {model_seed_id: $compound_id})-[s:is_intervenient_in {coef: $coef, role: $role}]->(r) "
                            "on create set s.timestamp = timestamp()",
                            coef = coef, role = role, compound_id = metabolite,
                            reaction_id = reaction_id)


            return reaction_id



    def generate_equation(self,onto_stoichiometry, ms_stoich, direction):

        left_side = []
        right_side = []

        for metabolite in onto_stoichiometry:

            coef = onto_stoichiometry[metabolite]

            if coef < 0:

                if coef < -1:

                    left_side.append(str(abs(coef)) + " C_BOIMMG_" + str(metabolite))

                else:

                    left_side.append("C_BOIMMG_" + str(metabolite))

            elif coef == 1:
                right_side.append("C_BOIMMG_" +str(metabolite))

            else:
                right_side.append(str(coef) + " C_BOIMMG_" + str(metabolite))

        for metabolite in ms_stoich:

            coef = ms_stoich[metabolite]

            if coef < 0:

                if coef < -1:

                    left_side.append(str(abs(coef)) + metabolite)

                else:

                    left_side.append(metabolite)

            elif coef == 1:
                right_side.append(metabolite)

            else:
                right_side.append(str(coef) + metabolite)

        left_side_str = "+".join(left_side)
        right_side_str = "+".join(right_side)

        equation = direction.join([left_side_str,right_side_str])

        return equation














import time

import cobra
from cobra import Reaction

from boimmgpy import definitions
from boimmgpy.boimmg.RepresentationRevisors.RepresentationRedundantCaseSolver import RepresentationRedundantCaseSolver
from boimmgpy.boimmg.RepresentationRevisors.RepresentationSimpleCaseSolver import SimpleCaseSolver

# ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
from boimmgpy.case_studies.redundant_representation_case.map_ecoli_model import map_iJR904


def simulate_ecoli_generalize():
    model = cobra.io.read_sbml_model(definitions.ROOT_DIR + "/models/iML1515.xml")
    rep_case_solver = SimpleCaseSolver(model, "BiGG")
    rep_case_solver.swap_compound("cpd15560", "cpd11669")
    rep_case_solver.swap_compound("cpd15500", "cpd11451")
    cobra.io.write_sbml_model(rep_case_solver.model, "simple_representation_case/iML1515_generalized.xml")

def simulate_ecoli_granulate():
    model = cobra.io.read_sbml_model(definitions.ROOT_DIR + "/case_studies/iML1515_generalized.xml")
    gapfiller = SimpleCaseSolver(model, "BiGG")
    gapfiller.swap_compound("cpd11669","cpd15290")
    gapfiller.swap_compound("cpd11451", "cpd15500")
    cobra.io.write_sbml_model(gapfiller.model, "simple_representation_case/iML1515_granulated.xml")

def simulate_ecoli_granulate_boimmg():
    model = cobra.io.read_sbml_model(definitions.ROOT_DIR + "/case_studies/iML1515_generalized.xml")
    gapfiller = SimpleCaseSolver(model, "BiGG")
    gapfiller.swap_compound("cpd11669", "C_BOIMMG_749196")
    # gapfiller.swap_compound("cpd11451", "cpd15500")
    cobra.io.write_sbml_model(gapfiller.model, "simple_representation_case/iML1515_granulated_ubi5.xml")

def simulate_sacc_generalize():
    model = cobra.io.read_sbml_model("C:/Users/Biosystems/Desktop/thesis_jcapela/BOIMMGpy/models/iMM904.xml")
    gapfiller = SimpleCaseSolver(model, "BiGG")
    gapfiller.swap_compound("cpd15290", "cpd11669")
    cobra.io.write_sbml_model(gapfiller.model, "simple_representation_case/iMM904_generalized.xml")

def simulate_sacc_granulate():
    model = cobra.io.read_sbml_model("/BOIMMGpy/case_studies/simple_representation_case/iMM904_generalized.xml")
    gapfiller = SimpleCaseSolver(model, "BiGG")
    gapfiller.swap_compound("cpd11669" , "cpd15290")
    cobra.io.write_sbml_model(gapfiller.model, "simple_representation_case/iMM904_granulated.xml")

def convert_into_json():

    model = cobra.io.read_sbml_model(
        definitions.ROOT_DIR + "/case_studies/iMM904_generalized.xml")

    cobra.io.save_json_model(model,"iMM904_generalized.json")
    # model2 = cobra.io.read_sbml_model(
    #     ROOT_DIR + "/BOIMMGpy/case_studies/iML1515_generalized.xml")
    #
    # model3 = cobra.io.read_sbml_model(
    #     ROOT_DIR + "/BOIMMGpy/case_studies/iML1515_generalized.xml")

def redundant_granulator_ecoli():
    model = map_iJR904()

    components = ["cpd00214", "cpd01080", "cpd03847", "cpd05274"]

    solver = RepresentationRedundantCaseSolver(model, "BiGG")

    solver.swap_from_generic(["cpd22513", "cpd15649"], components, True,True)


    cobra.io.write_sbml_model(solver.model, "enhanced_model_ecoli_latest_version.xml")
    cobra.io.save_json_model(solver.model, "enhanced_model_ecoli.json")

def redundant_granulator_ecoli_without_components():
    model = map_iJR904()

    components = ["cpd00214", "cpd03847", "cpd05274","cpd25615","cpd05237"]

    solver = RepresentationRedundantCaseSolver(model, "BiGG")

    solver.swap_from_generic(["cpd22513", "cpd15649"], components, True)
    solver.generateISAreactions()

    cobra.io.write_sbml_model(solver.model, "redundant_representation_case/enhanced_model_ecoli_without_components.xml")
    cobra.io.save_json_model(solver.model, "enhanced_model_ecoli.json")

def redundant_granulator_ecoli_without_components_isa_reactions_for_all():
    model = map_iJR904()

    components = ["cpd00214", "cpd03847", "cpd05274","cpd25615","cpd05237"]

    solver = RepresentationRedundantCaseSolver(model, "BiGG")

    solver.swap_from_generic(["cpd22513", "cpd15649"], components, True)
    solver.generateTestISAreactions()

    cobra.io.write_sbml_model(solver.model, "enhanced_model_ecoli_without_components_isa_reactions_for_all.xml")
    cobra.io.save_json_model(solver.model, "enhanced_model_ecoli.json")

def gap_fill_ecoli_model():
    model = cobra.io.read_sbml_model(definitions.ROOT_DIR + "/case_studies/enhanced_model_ecoli_without_components.xml")

    components = ["palmACP_c", "myrsACP_c", "hdeACP_c", "octeACP_c","tdeACP_c"]

    PASYN_EC(model)
    LPLIPA2(model)
    LPLIPA1(model)
    LPLIPA3(model)
    PLIPA1(model)


    # model.solver = "cplex"
    # gapfiller = GapFiller(model,exchange_reactions=True, demand_reactions=True,integer_threshold=1e-300)
    # reactions = gapfiller.fill(5)
    # print(reactions)

    # model.add_reactions(reactions[0])
    cobra.io.write_sbml_model(model,
        definitions.ROOT_DIR + "/case_studies/enhanced_model_ecoli_without_components_gap_filled.xml")

def PLIPA1(model):
    s1 = {model.metabolites.get_by_id("C_BOIMMG_427_c"): -1,
          model.metabolites.get_by_id("hdca_c"): 1,
          model.metabolites.get_by_id("C_BOIMMG_314_c"): 1}

    s2 = {model.metabolites.get_by_id("C_BOIMMG_7454_c"): -1,
          model.metabolites.get_by_id("hdcea_c"): 1,
          model.metabolites.get_by_id("C_BOIMMG_16363_c"): 1}

    s3 = {model.metabolites.get_by_id("C_BOIMMG_7408_c"): -1,
          model.metabolites.get_by_id("ocdcea_c"): 1,
          model.metabolites.get_by_id("C_BOIMMG_16317_c"): 1}

    s4 = {model.metabolites.get_by_id("C_BOIMMG_7474_c"): -1,
          model.metabolites.get_by_id("ttdca_c"): 1,
          model.metabolites.get_by_id("C_BOIMMG_452_c"): 1}

    s5 = {model.metabolites.get_by_id("C_BOIMMG_7456_c"): -1,
          model.metabolites.get_by_id("ttdcea_c"): 1,
          model.metabolites.get_by_id("C_BOIMMG_16364_c"): 1}

    r1 = Reaction("PLIPA1_1")
    r2 = Reaction("PLIPA1_2")
    r3 = Reaction("PLIPA1_3")
    r4 = Reaction("PLIPA1_4")
    r5 = Reaction("PLIPA1_5")

    s = [s1, s2, s3, s4, s5]
    r = [r1, r2, r3, r4, r5]

    for i in range(len(r)):
        si = s[i]

        ri = r[i]

        si[model.metabolites.get_by_id("h2o_c")] = -1
        si[model.metabolites.get_by_id("h_c")] = 1

        ri.add_metabolites(si)

    model.add_reactions(r)

def LPLIPA1(model):
    s1 = {model.metabolites.get_by_id("C_BOIMMG_314_c"): -1,
          model.metabolites.get_by_id("hdca_c"): 1}

    s2 = {model.metabolites.get_by_id("C_BOIMMG_16363_c"): -1,
          model.metabolites.get_by_id("hdcea_c"): 1}

    s3 = {model.metabolites.get_by_id("C_BOIMMG_16317_c"): -1,
          model.metabolites.get_by_id("ocdcea_c"): 1}

    s4 = {model.metabolites.get_by_id("C_BOIMMG_452_c"): -1,
          model.metabolites.get_by_id("ttdca_c"): 1}

    s5 = {model.metabolites.get_by_id("C_BOIMMG_16364_c"): -1,
          model.metabolites.get_by_id("ttdcea_c"): 1}

    r1 = Reaction("LPLIPA1_1")
    r2 = Reaction("LPLIPA1_2")
    r3 = Reaction("LPLIPA1_3")
    r4 = Reaction("LPLIPA1_4")
    r5 = Reaction("LPLIPA1_5")

    s = [s1, s2, s3, s4, s5]
    r = [r1, r2, r3, r4, r5]

    for i in range(len(r)):
        si = s[i]

        ri = r[i]

        si[model.metabolites.get_by_id("h2o_c")] = -1
        si[model.metabolites.get_by_id("h_c")] = 1
        si[model.metabolites.get_by_id("g3pg_c")] = 1

        ri.add_metabolites(si)

    model.add_reactions(r)

def LPLIPA3(model):
    s1 = {model.metabolites.get_by_id("C_BOIMMG_223_c"): -1,
          model.metabolites.get_by_id("hdca_c"): 1}

    s2 = {model.metabolites.get_by_id("C_BOIMMG_6942_c"): -1,
          model.metabolites.get_by_id("hdcea_c"): 1}

    s3 = {model.metabolites.get_by_id("C_BOIMMG_6896_c"): -1,
          model.metabolites.get_by_id("ocdcea_c"): 1}

    s4 = {model.metabolites.get_by_id("C_BOIMMG_319_c"): -1,
          model.metabolites.get_by_id("ttdca_c"): 1}

    s5 = {model.metabolites.get_by_id("C_BOIMMG_6943_c"): -1,
          model.metabolites.get_by_id("ttdcea_c"): 1}

    r1 = Reaction("LPLIPA3_1")
    r2 = Reaction("LPLIPA3_2")
    r3 = Reaction("LPLIPA3_3")
    r4 = Reaction("LPLIPA3_4")
    r5 = Reaction("LPLIPA3_5")

    s = [s1, s2, s3, s4, s5]
    r = [r1, r2, r3, r4, r5]

    for i in range(len(r)):
        si = s[i]

        ri = r[i]

        si[model.metabolites.get_by_id("h2o_c")] = -1
        si[model.metabolites.get_by_id("h_c")] = 1
        si[model.metabolites.get_by_id("g3pc_c")] = 1

        ri.add_metabolites(si)

    model.add_reactions(r)

def LPLIPA2(model):
    s1 = {model.metabolites.get_by_id("C_BOIMMG_291_c"): -1,
          model.metabolites.get_by_id("hdca_c"): 1}

    s2 = {model.metabolites.get_by_id("C_BOIMMG_27923_c"): -1,
          model.metabolites.get_by_id("hdcea_c"): 1}

    s3 = {model.metabolites.get_by_id("C_BOIMMG_27877_c"): -1,
          model.metabolites.get_by_id("ocdcea_c"): 1}

    s4 = {model.metabolites.get_by_id("C_BOIMMG_27943_c"): -1,
          model.metabolites.get_by_id("ttdca_c"): 1}

    s5 = {model.metabolites.get_by_id("C_BOIMMG_27925_c"): -1,
          model.metabolites.get_by_id("ttdcea_c"): 1}

    r1 = Reaction("LPLIPA2_1")
    r2 = Reaction("LPLIPA2_2")
    r3 = Reaction("LPLIPA2_3")
    r4 = Reaction("LPLIPA2_4")
    r5 = Reaction("LPLIPA2_5")

    s = [s1, s2, s3, s4, s5]
    r = [r1, r2, r3, r4, r5]

    for i in range(len(r)):
        si = s[i]

        ri = r[i]

        si[model.metabolites.get_by_id("h2o_c")] = -1
        si[model.metabolites.get_by_id("h_c")] = 1
        si[model.metabolites.get_by_id("g3pe_c")] = 1

        ri.add_metabolites(si)

    model.add_reactions(r)

def PASYN_EC(model):
    s1 = {model.metabolites.get_by_id("palmACP_c"): -1,
          model.metabolites.get_by_id("C_BOIMMG_423_c"): 1}

    s2 = {model.metabolites.get_by_id("myrsACP_c"): -1,
          model.metabolites.get_by_id("C_BOIMMG_38435_c"): 1}

    s3 = {model.metabolites.get_by_id("hdeACP_c"): -1,
          model.metabolites.get_by_id("C_BOIMMG_38416_c"): 1}

    s4 = {model.metabolites.get_by_id("octeACP_c"): -1,
          model.metabolites.get_by_id("C_BOIMMG_38371_c"): 1}

    s5 = {model.metabolites.get_by_id("tdeACP_c"): -1,
          model.metabolites.get_by_id("C_BOIMMG_38417_c"): 1}

    r1 = Reaction("reaction_1")
    r2 = Reaction("reaction_2")
    r3 = Reaction("reaction_3")
    r4 = Reaction("reaction_4")
    r5 = Reaction("reaction_5")

    s = [s1, s2, s3, s4, s5]
    r = [r1, r2, r3, r4, r5]

    for i in range(len(r)):
        si = s[i]

        ri = r[i]

        si[model.metabolites.get_by_id("glyc3p_c")] = -1
        si[model.metabolites.get_by_id("ACP_c")] = 2

        ri.add_metabolites(si)

    model.add_reactions(r)

def redundant_granulator_ecoli_all():
    start = time.time()
    model = map_iJR904()

    components = ["cpd00214", "cpd01080", "cpd03847", "cpd05274"]

    solver = RepresentationRedundantCaseSolver(model, "BiGG")

    solver.swap_from_generic(["cpd22513", "cpd15649"], components, False,True)

    cobra.io.write_sbml_model(solver.model, "enhanced_model_ecoli_all.xml")
    cobra.io.save_json_model(solver.model, "enhanced_model_ecoli_all.json")

    end=time.time()

    print("time")
    print(end-start)

if __name__ == "__main__":
    # simulate_ecoli_generalize()
    # simulate_ecoli_granulate()
    # redundant_granulator_ecoli()
    # redundant_granulator_ecoli_without_components()

    # redundant_granulator_ecoli_without_components_isa_reactions_for_all()
    # redundant_granulator_ecoli_without_components()
    gap_fill_ecoli_model()
    # redundant_granulator_ecoli_all()
    # simulate_ecoli_granulate_boimmg()
    # simulate_sacc_generalize()
    # simulate_sacc_granulate()
    # convert_into_json()




def read_conf_file(path):
    '''
    reading the configuration file
    It must be assembled in the following manner: key = configuration value
    It will return a dictionary in the following format: {workers : 3, limit : 100}
    '''
    res = {}

    with open(path) as file:
        lines = file.readlines()
        for line in lines:
            if '=' in line:
                prop = line.strip()
                l_str = prop.split('=')
                if len(l_str) >2:
                    l_str[1] = "=".join(l_str[1:])

                res[l_str[0].strip()] = l_str[1].strip()

    return res

def change_conf_file(path, dic):
    new_lines = []
    with open(path,"r") as file:
        lines = file.readlines()
        for line in lines:
            if '=' in line:
                prop = line.strip()
                l_str = prop.split('=')
                new_prop = dic[l_str[0].strip()]
                new_line = l_str[0]+"="+new_prop+"\n"
                new_lines.append(new_line)
            else:
                new_lines.append(line)

    with open(path,"w") as file:
        file.writelines(new_lines)


import re

from neo4j import GraphDatabase

from boimmgpy import definitions
from boimmgpy.database.containers.compound_node import CompoundNode
from boimmgpy.database.databases_babel import AliasesTransformer
from boimmgpy.database.interfaces.boimmg_database_accessor import BOIMMGDatabaseAccessor
from boimmgpy.utilities import file_utilities

def set_database_information(uri,user,password):

    uri_bool = False
    user_bool = False
    pass_bool = False

    with open(definitions.BOIMMG_DATABASE, "r") as f:
        lines = f.readlines()

    with open(definitions.BOIMMG_DATABASE, "w") as f:
        for line in lines:
            if re.match("^uri=",line):
                uri_bool = True
                f.write("uri=" + uri + "\n")
            elif re.match("^user=",line):
                user_bool = True
                f.write("user=" + user + "\n")
            elif re.match("^password=",line):
                pass_bool = True
                f.write("password=" + password + "\n")
            else:
                f.write(line)

        if not uri_bool:
            f.write("uri=" + uri + "\n")

        if not user_bool:
            f.write("user=" + user + "\n")

        if not pass_bool:
            f.write("password=" + password + "\n")

class CompoundsDBAccessor(BOIMMGDatabaseAccessor):

    def __init__(self):


        uri, user, password = self.read_config_file()

        self.__uri = uri
        self.__user = user
        self.__password = password

    def read_config_file(self):

        configs = file_utilities.read_conf_file(definitions.BOIMMG_DATABASE)

        if "uri" in configs.keys() and "user" in configs.keys() and "password" in configs.keys():

            uri = configs["uri"]
            user = configs["user"]
            password = configs["password"]

            return uri, user, password

        else:

            raise Exception("Please insert the required database information")

    def login(self):

        driver = GraphDatabase.driver(self.__uri, auth=(self.__user, self.__password), encrypted=False)
        self.tx = driver

    @property
    def tx(self):
        return self._tx

    @tx.setter
    def tx(self,tx):
        self._tx = tx


    def add_relationship(self,origin : int, target : int, type : str):
        self.login()
        with self.tx.session() as session:
            session.run("MATCH (c:Compound), (p:Compound)"
                        "WHERE c.boimmg_id = $target and p.boimmg_id = $origin "
                        "MERGE (c)<-[r:" + type+ "]-(p) "
                        "ON CREATE SET r.timestamp = timestamp() ",
                        origin = origin, target= target, type = type)

    def add_biosynthetic_relationship(self,origin : int, target : int, **kwargs):
        self.login()
        properties = []
        for key in kwargs:
            properties.append( key + ":'"+ kwargs.get(key)+"'")

        properties_cypher = ",".join(properties)

        with self.tx.session() as session:
            session.run("MATCH (c:Compound), (p:Compound)"
                    "WHERE c.boimmg_id = $target and p.boimmg_id = $origin "
                    "MERGE (c)<-[r:precursor_of {" + properties_cypher + "} ]-(p) "
                    "ON CREATE SET r.timestamp = timestamp()", origin = origin,target=target
                    )



    def get_predecessors_by_swiss_lipids_id(self, db_id) -> list:
        self.login()
        predecessors = []
        with self.tx.session() as session:
            result = session.run("MATCH (c:Compound)<-[r]-(p:Compound) "
                        "WHERE c.swiss_lipids_id = $db_id "
                        "RETURN p, p.boimmg_id as id",
                        db_id=db_id)


            data = result.data()
            for node in data:
                node_properties = node.get("p")
                node_id = node.get("id")
                other_aliases = self.get_all_aliases(node_id)
                node = CompoundNode(node_id,node_properties,other_aliases)
                predecessors.append(node)

        return predecessors

    def get_predecessors_by_model_seed_id(self, db_id:str) -> list:
        self.login()
        predecessors = []
        with self.tx.session() as session:
            result = session.run("MATCH (c:Compound)<-[r]-(p:Compound) "
                        "WHERE c.model_seed_id = $db_id "
                        "RETURN p, p.boimmg_id as id",
                        db_id=db_id)


            data = result.data()
            for node in data:
                node_properties = node.get("p")
                node_id = node.get("id")
                other_aliases = self.get_all_aliases(node_id)
                node = CompoundNode(node_id,node_properties,other_aliases)
                predecessors.append(node)

        return predecessors

    def get_node_from_model_seed_id(self, model_seed_id:str) -> CompoundNode:
        self.login()
        with self.tx.session() as session:
            result = session.run("MATCH (c:ModelSeedCompound)-[:is_db_link_of]->(d:Compound) "
                                 "USING INDEX c:ModelSeedCompound(model_seed_id) "
                                 "WHERE c.model_seed_id = $model_seed_id "
                                 "RETURN d, d.boimmg_id as id",
                                  model_seed_id=model_seed_id)

            data = result.single()
            if data:
                node_properties = data.get("d")
                node_id = data.get("id")
                other_aliases = self.get_all_aliases(node_id)
                node = CompoundNode(node_id, node_properties,other_aliases)

                return node

            else: return None

    def get_all_aliases(self, node_id):
        self.login()
        with self.tx.session() as session:
            result = session.run("MATCH (c)-[:is_db_link_of]->(d:Compound) "
                                 "WHERE d.boimmg_id = $node_id "
                                 "RETURN c",
                                 node_id=node_id)

            data = result.data()
            res=[]
            if data:
                for node in data:
                    node_properties = node.get("c")
                    res.append(node_properties)


            return res

    def get_node_id_from_model_seed_id(self, model_seed_id: str) -> int:
        self.login()
        with self.tx.session() as session:

            result = session.run("MATCH (c:ModelSeedCompound)-[:is_db_link_of]->(d:Compound) "
                                 "USING INDEX c:ModelSeedCompound(model_seed_id) "
                                 "WHERE c.model_seed_id = $model_seed_id "
                                 "RETURN d.boimmg_id as id",
                                 model_seed_id=model_seed_id)

            data = result.single()
            if data:
                node_id = data.get("id")

                return node_id

            else:
                return None

    def get_node_from_lipid_maps_id(self, lipid_maps_id: str) -> CompoundNode:
        self.login()
        with self.tx.session() as session:
            result = session.run("MATCH (c:Compound) "
                                 "WHERE c.lipidmaps_id = $lipid_maps_id "
                                 "RETURN c, c.boimmg_id as id",
                                  lipid_maps_id=lipid_maps_id)

            data = result.single()
            if data:
                node_properties = data.get("c")
                node_id = data.get("id")
                other_aliases = self.get_all_aliases(node_id)
                node = CompoundNode(node_id, node_properties,other_aliases)

                return node

            else: return None

    def get_node_by_alias(self, key, alias) -> int:
        self.login()

        boimmg_property = AliasesTransformer.convert_model_seed_format_into_boimmg(key)

        if boimmg_property:
            with self.tx.session() as session:
                # result = session.run("MATCH (c:Compound) "
                #                      "WHERE c." + boimmg_property + "=$alias "
                #                      "RETURN ID(c) as id",
                #                      alias=alias)

                result = session.run("MATCH (c:Compound {" + boimmg_property + ": $alias}) "
                                     "RETURN c.boimmg_id as id",
                                     alias=alias)

                data = result.single()

                if data:
                    # node_properties = data.get("c")
                    node_id = data.get("id")
                    # node = CompoundNode(node_id, node_properties)

                    return node_id

        return

    def get_conjugates(self,ont_id : int) -> list:
        """Get conjugated acids and bases predecessors using as parameter the ontology identifier"""

        res = []
        conjugated_bases = \
            self.get_predecessors_by_ont_id_rel_type(ont_id,"conjugated_base_of")

        conjugated_acids =  \
            self.get_predecessors_by_ont_id_rel_type(ont_id,"conjugated_acid_of")

        res.extend(conjugated_bases)
        res.extend(conjugated_acids)

        conjugated_bases = \
            self.get_successors_by_ont_id_rel_type(ont_id, "conjugated_base_of")

        conjugated_acids = \
            self.get_successors_by_ont_id_rel_type(ont_id, "conjugated_acid_of")

        res.extend(conjugated_bases)
        res.extend(conjugated_acids)

        return res



    def get_predecessors_by_ont_id(self, ont_id: int) -> list:
        """Get predecessors using as parameter the ontology identifier"""
        self.login()
        predecessors = []
        with self.tx.session() as session:
            result = session.run("MATCH (c:Compound)<-[r]-(p:Compound) "
                                 "WHERE c.boimmg_id = $ont_id "
                                 "RETURN p.boimmg_id as id",
                                 ont_id=ont_id)

            data = result.data()
            for node in data:
                node_id = node.get("id")
                predecessors.append(node_id)

        return predecessors

    def get_predecessors_only_from_lipid_maps_and_ms(self,ont_id: int):
        self.login()
        predecessors = []

        with self.tx.session() as session:
            result = session.run(
                "match (c:Compound)<-[:is_a]-(d:Compound) where c.boimmg_id = $ont_id "
                "and not exists(d.swiss_lipids_id) and (not exists(d.annotated) or d.annotated) "
                "return d.boimmg_id as id",
                ont_id=ont_id)

            data = result.data()
            for node in data:
                node_id = node.get("id")
                predecessors.append(node_id)

        return predecessors


    def get_predecessors_with_same_component(self, ont_id: int, relationship_type : str) -> list:
        """Get predecessors with only one component using as parameters the database identifier and the relationship type"""
        self.login()
        predecessors = []
        with self.tx.session() as session:
            result = session.run("match (c:Compound)<-[:"+relationship_type+"]-(d:Compound)<-[:component_of]-(g) where c.boimmg_id = $ont_id "
                                "with collect(g) as components, d ,c "
                                 "where size(components) = 1 return d.boimmg_id as id",
                                 ont_id=ont_id,
                                 rel_type =relationship_type)

            data = result.data()
            if data:
                for node in data:
                    node_id = node.get("id")
                    predecessors.append(node_id)

        return predecessors

    def get_compounds_with_only_one_component(self, ont_id: int, components_ids_list: list) -> list:
        """Get predecessors with only one component using as parameter the database identifier and the relationship type"""
        self.login()
        predecessors = []
        with self.tx.session() as session:
            result = session.run("match (c:Compound)<-[:is_a]-(d:Compound)<-[:component_of]-(g) where c.boimmg_id = $ont_id "
                                "with collect(g) as components, d "
                                "where size(components) = 1 with d as result "
                                "match (result)<-[:component_of]-(t) where t.boimmg_id in $components_ids_list "
                                "return result.boimmg_id as id ",
                                 ont_id=ont_id,
                                 components_ids_list =components_ids_list)

            data = result.data()
            if data:
                for node in data:
                    node_id = node.get("id")
                    predecessors.append(node_id)

        return predecessors

    def get_all_model_seed_ids(self,ont_id):
        self.login()
        with self.tx.session() as session:
            result = session.run("MATCH (c:Compound)<-[r:is_db_link_of]-(p:ModelSeedCompound) "
                                 "WHERE c.boimmg_id = $ont_id "
                                 "RETURN p.model_seed_id as model_seed_id",
                                 ont_id=ont_id)

            data = result.data()

            model_seed_ids = []
            if data:
                for node in data:
                    model_seed_id = node.get("model_seed_id")
                    model_seed_ids.append(model_seed_id)

        return model_seed_ids


    def get_predecessors_by_ont_id_rel_type(self, ont_id: int, relationship_type : str) -> list:
        """Get predecessors using as parameter the database identifier and the relationship type"""
        self.login()
        predecessors = []
        with self.tx.session() as session:
            result = session.run("MATCH (c:Compound)<-[r]-(p:Compound) "
                                 "WHERE c.boimmg_id = $ont_id AND TYPE(r) = $rel_type "
                                 "RETURN p.boimmg_id as id",
                                 ont_id=ont_id,
                                 rel_type =relationship_type)

            data = result.data()
            if data:
                for node in data:
                    node_id = node.get("id")
                    predecessors.append(node_id)

        return predecessors

    def get_predecessors_by_ont_id_reaction_id(self, ont_id: int, reaction_id: str) -> list:
        """Get predecessors using as parameter the database identifier and the relationship type"""
        self.login()
        predecessors = []
        with self.tx.session() as session:
            result = session.run("MATCH (c:Compound)<-[r:precursor_of]-(p:Compound) "
                                 "WHERE c.boimmg_id = $ont_id AND r.reaction = $reaction_id "
                                 "RETURN p.boimmg_id as id",
                                 ont_id=ont_id,
                                 reaction_id =reaction_id)

            data = result.data()
            if data:
                for node in data:
                    node_id = node.get("id")
                    predecessors.append(node_id)

        return predecessors

    def get_all_successors_by_ont_id_rel_type(self,ont_id: int, relationship_type : str,threshold=50) -> list:
        self.login()

        parent_container = self.get_node_by_ont_id(ont_id)
        l = [ont_id]
        res = []
        while len(l) > 0:
            node = l.pop(0)
            successors = self.get_successors_by_ont_id_rel_type(node, relationship_type)
            if node != parent_container.id:
                res.append(node)
            for elem in successors:

                if elem not in res and elem not in l:
                    l.append(elem)
        return res

    def get_all_predecessors_by_ont_id_rel_type(self, ont_id: int, relationship_type : str) -> list:
        self.login()

        parent_container = self.get_node_by_ont_id(ont_id)
        l = [ont_id]
        res = []
        while len(l) > 0:
            node = l.pop(0)
            predecessors = self.get_predecessors_by_ont_id_rel_type(node, relationship_type)
            if node != parent_container.id:
                res.append(node)
            for elem in predecessors:
                if elem not in res and elem not in l:
                    l.append(elem)
        return res


    def get_compound_by_inchikey(self,inchikey):
        self.login()
        with self.tx.session() as session:
            result = session.run("MATCH (c:Compound) "
                                 "WHERE c.inchikey contains $inchikey "
                                 "RETURN c, c.boimmg_id as id "
                                 ,
                                 inchikey=inchikey[:-1])
            data = result.single()
            if data:

                node_properties = data.get("c")
                node_id = data.get("id")
                other_aliases = self.get_all_aliases(node_id)
                node = CompoundNode(node_id, node_properties,other_aliases)
                return node

            else: return None

    def get_compound_by_smiles_and_abstraction(self,smiles,generic):
        self.login()
        with self.tx.session() as session:
            result = session.run("MATCH (c:Compound) "
                                 "WHERE c.smiles = $smiles and c.generic = $generic "
                                 "RETURN c, c.boimmg_id as id "
                                 ,
                                 smiles=smiles,
                                 generic = generic)

            data = result.single()
            if data:

                node_properties = data.get("c")
                node_id = data.get("id")
                other_aliases = self.get_all_aliases(node_id)
                node = CompoundNode(node_id, node_properties,other_aliases)
                return node

            else:
                return None

    def get_compound_by_smiles(self,smiles):
        self.login()
        with self.tx.session() as session:
            result = session.run("MATCH (c:Compound) "
                                 "WHERE c.smiles = $smiles "
                                 "RETURN c, c.boimmg_id as id "
                                 ,
                                 smiles = smiles)
            data = result.single()
            if data:

                node_properties = data.get("c")
                node_id = data.get("id")
                other_aliases = self.get_all_aliases(node_id)
                node = CompoundNode(node_id, node_properties,other_aliases)
                return node

            else:
                return None


    def create_compound(self,**kwargs) -> CompoundNode:

        name = kwargs.get("name")
        formula = kwargs.get("formula")
        smiles = kwargs.get("smiles")
        generic = bool(kwargs.get("generic"))
        charge = kwargs.get("charge")
        inchikey = kwargs.get("inchikey")
        annotated = kwargs.get("annotated")
        inchi = kwargs.get("inchi")
        aliases = kwargs.get("aliases")
        self.login()
        with self.tx.session() as session:
            result = session.run("match (c:Compound) return MAX(c.boimmg_id) as max_id")
            data = result.single()

            last_id = data.get("max_id")

            new_id = last_id +1

        if aliases:
            return self.__create_compound_with_aliases_boimmg_format(boimmg_id = new_id, name = name, formula = formula,
                                                                     smiles = smiles,
                                                                     generic = generic,
                                                                     charge = charge,
                                                                     inchikey = inchikey, annotated = annotated,
                                                                     inchi = inchi, aliases = aliases)

        else:
            self.login()
            with self.tx.session() as session:
                result = session.run("MERGE (c:Compound {name : $name,"
                            "formula : $formula, smiles: $smiles,"
                            "generic: $generic ,"
                            "charge: $charge, inchikey:$inchikey, "
                            "annotated:$annotated }) "
                            "on create set c.boimmg_id = $id "
                            "RETURN c, c.boimmg_id as id"
                            ,
                             name = name, formula = formula, smiles = smiles,
                            generic = generic, charge = charge, inchikey = inchikey,
                            annotated=annotated,id = new_id)


                data = result.single()
                node_properties = data.get("c")
                node_id = data.get("id")
                other_aliases = self.get_all_aliases(node_id)
                node = CompoundNode(node_id, node_properties,other_aliases)

                return node

    def get_node_by_ont_id(self, ont_id: int) -> CompoundNode:
        """Get node container by ontology identifier"""
        self.login()
        with self.tx.session() as session:
            result = session.run("MATCH (c:Compound) "
                                 "WHERE c.boimmg_id = $ont_id "
                                 "RETURN c, c.boimmg_id as id",
                                 ont_id=ont_id)

            data = result.single()
            node_properties = data.get("c")
            node_id = data.get("id")

            other_aliases = self.get_all_aliases(node_id)
            node = CompoundNode(node_id, node_properties,other_aliases)

        return node

    def get_compounds_with_specific_parent_within_set_of_components(self, parent, components):

        self.login()
        with self.tx.session() as session:
            result = session.run("match (e:Compound)<-[:is_a]-(c:Compound)<-[:component_of]-(d:Compound) "
                                 "with collect(d) as components,e,c "
                                 "where e.boimmg_id = $parent and "
                                 "all (x in components where x.boimmg_id in $components_par) "
                                 "return c.boimmg_id as id ",
                                 components_par=components,
                                 parent = parent
                                 )

            data = result.data()
            res = []
            for node in data:
                node_id = node.get("id")
                res.append(node_id)

        return res

    def get_compounds_with_specific_parent_set_of_components(self, parent, components):

        self.login()
        with self.tx.session() as session:
            result = session.run("match (e:Compound)<-[:is_a]-(c:Compound)<-[:component_of]-(d:Compound) "
                                 "with collect(id(d)) as components,e,c "
                                 "where e.boimmg_id = $parent and "
                                 "components = $components_par "
                                 "return c.boimmg_id as id ",
                                 components_par=components,
                                 parent=parent
                                 )

            data = result.data()
            res = []
            for node in data:
                node_id = node.get("id")
                res.append(node_id)

        return res

        # res=[]
        # for child in children:
        #     child_components = self.get_predecessors_by_ont_id_rel_type(child,"component_of")
        #
        #     inList = True
        #     if child_components:
        #         for component in child_components:
        #             if component not in components:
        #                 inList = False
        #                 break
        #
        #         if inList:
        #             res.append(child)
        # return res



    def get_leaves_from_ont_id(self,ont_id:int) -> list:

        parent_container = self.get_node_by_ont_id(ont_id)
        if parent_container.generic:
            l = [parent_container.id]
            res = []
            while len(l) > 0:
                node = l.pop(0)
                predecessors = self.get_predecessors_by_ont_id_rel_type(node, "is_a")
                if node != parent_container.id:
                    if not predecessors:
                        res.append(node)
                for elem in predecessors:
                    if elem not in res and elem not in l:
                        l.append(elem)
            return res
        else:
            return []

    def get_all_parents(self,leaf):

        l = [leaf]
        res = []
        while len(l) > 0:

            node = l.pop(0)
            successors = self.get_successors_by_ont_id_rel_type(node, "is_a")

            if node != leaf:
                res.append(node)

            for elem in successors:
                if elem not in res and elem not in l:
                    l.append(elem)

        return res

    def check_if_node_exist_by_model_seed_id(self,model_seed_id):
        self.login()
        with self.tx.session() as session:
            result = session.run("MATCH (c:Compound) "
                                 "WHERE c.model_seed_id = $model_seed_id "
                                 "RETURN c, c.boimmg_id as id",
                                 model_seed_id=model_seed_id)

            data = result.single()
            node = None
            if data:
                node_properties = data.get("c")
                node_id = data.get("id")

                other_aliases = self.get_all_aliases(node_id)
                node = CompoundNode(node_id, node_properties,other_aliases)

        return node

    def get_successors_by_ont_id_rel_type(self, ont_id:int, relationship_type:str) -> list:
        self.login()
        successors = []
        with self.tx.session() as session:

            result = session.run("MATCH (c:Compound)<-[r]-(p:Compound) "
                                 "WHERE p.boimmg_id = $ont_id AND TYPE(r) = $rel_type "
                                 "RETURN c.boimmg_id as id",
                                 ont_id=ont_id,
                                 rel_type=relationship_type)

            data = result.data()
            if data:
                for node in data:
                    node_id = node.get("id")
                    successors.append(node_id)
        return successors

    def get_successors_and_biosynthetic_relationship_info_by_ont_id_rel_type(self, ont_id:int, relationship_type:str):
        self.login()
        successors = []
        with self.tx.session() as session:
            result = session.run("MATCH (c:Compound)<-[r]-(p:Compound) "
                                 "WHERE p.boimmg_id = $ont_id AND TYPE(r) = $rel_type "
                                 "return r.pathway as pathway, r.reaction as reaction, c.boimmg_id as id ",
                                 ont_id = ont_id, rel_type=relationship_type)

            data = result.data()
            if data:
                for node in data:
                    node_id = node.get("id")
                    pathway = node.get("pathway")
                    reaction = node.get("reaction")
                    successors.append((node_id,pathway,reaction))

        return successors



    def establish_structural_relationship(self,origin,target):
        self.login()

        with self.tx.session() as session:
            session.run("MATCH (c:Compound),(p:Compound) " 
                        "where c.boimmg_id = $target and p.boimmg_id = $origin "
                        "MERGE (c)<-[r:is_a]-(p) "
                         "on create set r.time_stamp = TIMESTAMP()",
                        target = target, origin = origin)

    def __create_compound_with_aliases_boimmg_format(self,**kwargs):

        name = kwargs.get("name")
        formula = kwargs.get("formula")
        smiles = kwargs.get("smiles")
        generic = kwargs.get("generic")
        charge = kwargs.get("charge")
        inchikey = kwargs.get("inchikey")
        annotated = kwargs.get("annotated")
        inchi = kwargs.get("inchi")
        aliases = kwargs.get("aliases")
        boimmg_id = kwargs.get("boimmg_id")


        self.login()
        with self.tx.session() as session:
            result = session.run("MERGE (c:Compound {name : $name, inchikey:$inchikey}) "
                                 "ON create set "
                                 "c.boimmg_id = $boimmg_id, "
                                 "c.formula = $formula, c.smiles = $smiles,"
                                 "c.generic = $generic ,"
                                 "c.charge = $charge,"
                                 "c.annotated = $annotated , c.inchi = $inchi "
                                 "RETURN c, c.boimmg_id as id"
                                 ,
                                 name=name, formula=formula, smiles=smiles,
                                 generic=generic, charge=charge, inchikey=inchikey,
                                 annotated=annotated, inchi = inchi, boimmg_id=boimmg_id)

            data = result.single()
            node_properties = data.get("c")
            node_id = data.get("id")
            for alias in aliases:
                session.run("MATCH (c:Compound) "
                            "where c.boimmg_id = $id "
                            "SET c."+alias+" = $alias_value ",
                            alias_value = aliases[alias],
                            id = node_id)

            other_aliases = self.get_all_aliases(node_id)
            node = CompoundNode(node_id, node_properties,other_aliases)

            return node







from boimmgpy.utilities import file_utilities
from boimmgpy.definitions import COMPOUNDS_ANNOTATION_CONFIGS_PATH,REACTIONS_ANNOTATION_CONFIGS_PATH

class AnnotationUtils(object):


    @staticmethod
    def get_compound_annotation_format_by_aliases(aliases):
        compoundsAnnotationConfigs = file_utilities.read_conf_file(COMPOUNDS_ANNOTATION_CONFIGS_PATH)

        res = {}
        for alias in aliases:
            if alias in compoundsAnnotationConfigs.keys():
                new_alias = compoundsAnnotationConfigs[alias]
                res[new_alias] = aliases[alias]
        return res

    @staticmethod
    def get_reaction_annotation_format_by_aliases(aliases):
        reactionsAnnotationConfigs = file_utilities.read_conf_file(
            REACTIONS_ANNOTATION_CONFIGS_PATH)

        res = {}
        for alias in aliases:
            if alias in reactionsAnnotationConfigs.keys():
                new_alias = reactionsAnnotationConfigs[alias]
                res[new_alias] = aliases[alias]
        return res

    @staticmethod
    def get_annotation_from_cobra_annotation(compound):
        compoundsAnnotationConfigs = file_utilities.read_conf_file(COMPOUNDS_ANNOTATION_CONFIGS_PATH)
        inverse = {value: key for key, value in compoundsAnnotationConfigs.items()}

        res = {}
        for annotation_key in compound.annotation:
            if annotation_key in inverse.keys():
                boimmg_database = inverse[annotation_key]

                aliases = compound.annotation[annotation_key]

                if isinstance(aliases,list):
                    res[boimmg_database] = aliases
                else:
                    res[boimmg_database] = [aliases]

        return res


class ModelSeedCompound:

    def __init__(self,properties):

        self.__id = properties.get("modelseed_id")
        self.__name = properties.get("name")
        self.__smiles = properties.get("smiles")
        self.__inchikey = properties.get("inchikey")
        self.__formula = properties.get("formula")
        self.__charge = properties.get("charge")
        self.__inchi = properties.get("inchi")

    def setStructures(self,smiles=None,inchikey=None,inchi = None):
        if smiles is not None:
            self.__smiles=smiles
        if inchikey is not None:
            self.__inchikey=inchikey

        if inchi:
            self.__inchi = inchi

    def getDbId(self):
        return self.__id

    def getName(self):
        return self.__name

    def getSmiles(self):
        return self.__smiles

    def getInchikey(self):
        return self.__inchikey

    def getFormula(self):
        return self.__formula

    def getCharge(self):
        return self.__charge

    def getInchi(self):
        return self.__inchi




import logging
from logging.handlers import TimedRotatingFileHandler

from cobra import Model, Metabolite

from boimmgpy.database.containers.compound_node import CompoundNode
from boimmgpy.database.accessors.compounds_database_accessor import CompoundsDBAccessor
from boimmgpy.model_seed.model_seed_compound import ModelSeedCompound
from boimmgpy.utilities import file_utilities, model_utilities
from boimmgpy.utilities.annotation_utils import AnnotationUtils
from boimmgpy.service.network_modifiers.reactions_swapper import ReactionsChanger
from boimmgpy.id_converters.compounds_id_converter import CompoundsIDConverter
from boimmgpy.kegg.kegg_compound import KeggCompound
from cobrababel import bigg
from boimmgpy.definitions import ROOT_DIR,TOOL_CONFIG_PATH,REACTIONS_ANNOTATION_CONFIGS_PATH,COMPOUNDS_ANNOTATION_CONFIGS_PATH

logPath = ROOT_DIR + "/logs"

# format the log entries, backup limit and others
formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s %(message)s')
handler = TimedRotatingFileHandler(logPath, when='midnight', backupCount=20)
handler.setFormatter(formatter)
logger = logging.getLogger(__name__)
logger.addHandler(handler)
logger.setLevel(logging.DEBUG)



class MetaboliteSwapper:

    def __init__(self, model, metabolite_in_model, new_metabolite, type,
                 modelSeedCompoundsDb, db_accessor,
                 model_database="ModelSEED",
                 compoundsIdConverter=None,
                 universal_model = None,
                 not_to_change_compounds = []):

        """
       Class constructor

       :param Model model: COBRApy model
       :param int metabolite_in_model: ontology id of the metabolite to be replaced
       :param int new_metabolite: ontology id of the metabolite that will replace
       :param int type: type of change (0, 1, 2 or 3)
       :param ModelSeedDB modelSeedCompoundsDb: Model SEED compounds database
       :param BOIMMGDatabaseAccessor: accessor (either raw or restAPI)
       :param string model_database: database format of metabolites and reactions
       :param (optional) CompoundsIDConverter compoundsIdConverter: compound id converter
       :param (optional) Model universal_model: an universal model
       :param (optional) list not_to_change_compounds: list with compound that are not supposed to be changed
       """


        if not universal_model:
            self.__universal_model = universal_model

        else:
            self.__universal_model = Model("universal_reactions")


        if not compoundsIdConverter:
            self.__compoundsIdConverter = CompoundsIDConverter()
        else:
            self.__compoundsIdConverter = compoundsIdConverter

        self.__compounds_ontology = db_accessor

        self.__model_database = model_database
        self.__compounds_db = modelSeedCompoundsDb
        self.__type = type
        self.__metabolite_in_model = metabolite_in_model
        self.__new_metabolite = new_metabolite
        self.__model = model
        self.__metabolites_not_to_change = not_to_change_compounds

        self.define_instance_variables()

    @property
    def model(self):
        return self.__model

    @model.setter
    def model(self,value):
        self.__model = value

    @property
    def mapper(self):
        return self.__mapper

    @mapper.setter
    def mapper(self,value):
        self.__mapper = value

    @property
    def changed_reactions(self):
        return self._changed_reactions

    @changed_reactions.setter
    def changed_reactions(self,value):
        self._changed_reactions = value


    def define_instance_variables(self):

        """
        Define instance variables

        :return:
        """

        self.report_material = {}
        self.__compoundsAnnotationConfigs = file_utilities.read_conf_file(
            COMPOUNDS_ANNOTATION_CONFIGS_PATH)
        self.__reactionsAnnotationConfigs = file_utilities.read_conf_file(
            REACTIONS_ANNOTATION_CONFIGS_PATH)

        self.__home_path__ = ROOT_DIR
        self.__configs = file_utilities.read_conf_file(
            TOOL_CONFIG_PATH)

        self.changed_reactions = []
        self.reactions_swapper = ReactionsChanger(self.model,
                                                  self.__type, self.__model_database, self.__universal_model,
                                                  compounds_converter=self.__compoundsIdConverter)

    def set_type(self, new_type):
        """
        Method to change the type of swap.
        :param int new_type: new type of swap
        :return:
        """
        self.__type = new_type
        self.reactions_swapper.set_type(new_type)

    def set_new_compound(self, new_compound):
        """
        Method to set the new compound that will replace the other in the model.
        :param int new_compound: ontology identifier of the new compound.
        :return:
        """
        self.__new_metabolite = new_compound

    def set_compound_in_model(self, compound_in_model):
        """
        Method to set the compound to be replaced.
        :param int compound_in_model: ontology identifier of the compound in the model.
        :return:
        """
        self.__metabolite_in_model = compound_in_model

    def get_universal_model(self):
        """
        Get the instance variable __universal_model.
        :return: Model __universal_model: return the universal model.
        """
        return self.__universal_model

    def swap_metabolites(self) -> list:
        """
        Algorithm to swap metabolites:
            1º Set the game changer: dictionary with the metabolites' intermediates, conjugated acid or/and base.
                - {Metabolite in model : metabolite replacer}
            2º Swap all the metabolite's intermediates, conjugated acid or/and base and the metabolite itself.
            3º Swap the reactions associated to the replaced metabolite using the ReactionSwapper class.

        :return list : changed reactions
        """

        compounds_with_reactions_to_change=[]
        game_changer = self.__get_game_changer()

        game_changer_in_model = {}
        for compound in game_changer.keys():
            compound_container = self.__compounds_ontology.get_node_by_ont_id(compound)
            in_model_compound_inchikey = compound_container.inchikey

            modelseed_id = compound_container.model_seed_id
            in_model_aliases={}
            if modelseed_id:
                in_model_aliases = self.__compoundsIdConverter.get_all_aliases_by_modelSeedID(modelseed_id)

            metabolitesInModel = self.mapper.check_metabolites_in_model(in_model_compound_inchikey,
                                                                        in_model_aliases, boimmg_container=compound_container)

            if metabolitesInModel:
                for metabolite in metabolitesInModel:
                    game_changer_in_model[metabolite] = game_changer[compound]

        for metabolite in game_changer_in_model.keys():

            inModel = game_changer_in_model[metabolite][0]
            if not inModel:
                new_compound_ontology_id = game_changer_in_model[metabolite][1]

                if self.__model_database == "ModelSEED":
                    self.__swap_modelseed_compound(metabolite, new_compound_ontology_id)

                elif self.__model_database == "BiGG":
                    self.__swap_bigg_metabolite(metabolite, new_compound_ontology_id)

                elif self.__model_database == "KEGG":
                    self.__swap_kegg_metabolite(metabolite, new_compound_ontology_id)

                compounds_with_reactions_to_change.append(metabolite)

            else:
                mets = game_changer_in_model[metabolite][1]
                for met in mets:
                    if metabolite.compartment == met.compartment:
                        self.change_reaction_by_swapping_existing_metabolites(metabolite,met)
                        compounds_with_reactions_to_change.append(met)

        changed_reactions = self.reactions_swapper.swap_reactions_by_compounds(compounds_with_reactions_to_change)
        return changed_reactions


    def __swap_conjugates(self,model_metabolite = None,new_metabolite = None):
        """
        Method to swap the conjugated base and acid of the compounds being swapped.
        If the model_metabolite and the new_metabolite are provided, the method will replace directly the model_metabolite
        for the new_metabolite.

        :param (optional) Metabolite model_metabolite: metabolite in the model to be replaced
        :param (optional) Metabolite new_metabolite:  metabolite to replace.
        :return:
        """
        conjugates_game_changer = self.__get_game_changer_conjugates()

        if not new_metabolite and not model_metabolite:

            compounds_with_reactions_to_change = []

            game_changer_in_model = {}
            for compound in conjugates_game_changer.keys():
                compound_container = self.__compounds_ontology.get_node_by_ont_id(compound)
                in_model_compound_inchikey = compound_container.inchikey
                in_model_aliases = compound_container.aliases
                metabolitesInModel = self.mapper.check_metabolites_in_model(in_model_compound_inchikey,
                                                                               in_model_aliases)
                if metabolitesInModel:
                    for metabolite_to_be_changed in metabolitesInModel:
                        game_changer_in_model[metabolite_to_be_changed] = conjugates_game_changer[compound]

            for metabolite_to_be_changed in game_changer_in_model.keys():

                inModel = game_changer_in_model[metabolite_to_be_changed][0]
                if not inModel:
                    new_compound_ontology_id = game_changer_in_model[metabolite_to_be_changed][1]

                    if self.__model_database == "ModelSEED":
                        self.__swap_modelseed_compound(metabolite_to_be_changed, new_compound_ontology_id)

                    elif self.__model_database == "BiGG":
                        self.__swap_bigg_metabolite(metabolite_to_be_changed, new_compound_ontology_id)

                    elif self.__model_database == "KEGG":
                        self.__swap_kegg_metabolite(metabolite_to_be_changed, new_compound_ontology_id)

                    compounds_with_reactions_to_change.append(metabolite_to_be_changed)

            self.reactions_swapper.swap_reactions_by_compounds(compounds_with_reactions_to_change)

        else:
            reactions_to_change = []
            for compound in conjugates_game_changer.keys():
                conjugates = conjugates_game_changer[compound][1]
                for conjugate in conjugates:
                    if conjugate.compartment == new_metabolite.compartment:

                        compound_container = self.__compounds_ontology.get_node_by_ont_id(compound)

                        in_model_compound_inchikey = compound_container.inchikey
                        in_model_aliases = compound_container.aliases

                        metabolitesInModel = self.mapper.check_metabolites_in_model(in_model_compound_inchikey,
                                                                             in_model_aliases,
                                                                             compound_container)

                        for metabolite_to_be_changed in metabolitesInModel:
                            if metabolite_to_be_changed.compartment == new_metabolite.compartment:

                                reactions_to_change = self.change_reaction_by_swapping_existing_metabolites(
                                   metabolite_to_be_changed , conjugate)

                    else:
                        db_type = self.__compoundsAnnotationConfigs.get(self.__model_database)
                        db_id = conjugate.annotation.get(db_type)

                        if type(db_id) == str:
                            modelseed_id = self.__compoundsIdConverter.convert_db_id_to_model_seed_by_db_id(db_id)
                        else:
                            modelseed_id = self.__compoundsIdConverter.convert_db_id_to_model_seed_by_db_id(db_id[0])


                        cont = self.__compounds_db.get_compound_by_id(modelseed_id[0])
                        new_compound = self.generate_new_metabolites(cont)

                        compound_ont_container = self.__compounds_ontology.get_node_by_ont_id(compound)

                        compound_to_be_changed = \
                            self.mapper.check_metabolites_in_model(compound_ont_container.inchikey,
                                                            compound_ont_container.aliases,
                                                            compound_ont_container)

                        if compound_to_be_changed:

                            reactions_to_change = self.change_reaction_by_swapping_existing_metabolites(
                                compound_to_be_changed[0], new_compound[0])

            if reactions_to_change:
                self.reactions_swapper.swap_reactions(reactions_to_change)


    def swap_only_metabolites_and_conjugates(self,model_metabolite = None,new_metabolite = None):
        """
        Method to swap only the metabolites and their conjugated acid and/or base. This method does not change the
        metabolite intermediates/precursors.

        :param (optional) Model.Metabolite model_metabolite: metabolite in the model to be replaced
        :param (optional) Model.Metabolite new_metabolite:  metabolite to replace.
        :return:
        """

        compounds_with_reactions_to_change = []

        if not model_metabolite:
            self.__swap_conjugates()


            ontology_container = self.__compounds_ontology.get_node_by_ont_id(self.__metabolite_in_model)
            inchikey = ontology_container.inchikey
            aliases = self.__compoundsIdConverter.get_all_aliases_by_modelSeedID(ontology_container.model_seed_id)
            model_metabolites = self.mapper.check_metabolites_in_model(inchikey,aliases)

            replacer_metabolites_in_model = \
                self.__check_if_compound_exists_in_model_by_ontology_id(self.__new_metabolite)

            if not replacer_metabolites_in_model:

                if self.__model_database == "ModelSEED":
                    for model_metabolite in model_metabolites:
                        self.__swap_modelseed_compound(model_metabolite, self.__new_metabolite)

                elif self.__model_database == "BiGG":
                    for model_metabolite in model_metabolites:
                        self.__swap_bigg_metabolite(model_metabolite, self.__new_metabolite)

                elif self.__model_database == "KEGG":
                    for model_metabolite in model_metabolites:
                        self.__swap_kegg_metabolite(model_metabolite, self.__new_metabolite)

                compounds_with_reactions_to_change.append(model_metabolite)

                self.reactions_swapper.swap_reactions_by_compounds(compounds_with_reactions_to_change)

            else:

                for model_metabolite in model_metabolites:
                    for replacer_metabolite in replacer_metabolites_in_model:
                        if model_metabolite.compartment == replacer_metabolite.compartment:

                            reactions_to_change = self.change_reaction_by_swapping_existing_metabolites(
                                model_metabolite, replacer_metabolite)

                            self.reactions_swapper.swap_reactions(reactions_to_change)

        elif new_metabolite:
            self.__swap_conjugates(model_metabolite,new_metabolite)

            reactions_to_change = self.change_reaction_by_swapping_existing_metabolites(
                model_metabolite,new_metabolite)

            changed_reactions = self.reactions_swapper.swap_reactions(reactions_to_change)
            self.changed_reactions.extend(changed_reactions)



    def change_reaction_by_swapping_existing_metabolites(self, model_metabolite, new_metabolite):
        """
        Method to change a given metabolite and the respective reactions.

        :param Model.Metabolite model_metabolite: metabolite in the model to be replaced
        :param Model.Metabolite new_metabolite:  metabolite to replace.
        :return: list<Reaction> reactions_to_change: list with the reactions to be changed (format, id, annotation, etc...)
        """

        reactions_to_change = []
        if model_metabolite.id != new_metabolite.id:

            reactions = model_metabolite.reactions
            for reaction in reactions:
                met_to_add = {}
                met_to_subtract = {}
                stoichiometry = reaction.metabolites
                coef = stoichiometry.get(model_metabolite)
                met_to_add[new_metabolite] = coef
                met_to_subtract[model_metabolite] = coef
                reaction.subtract_metabolites(met_to_subtract)
                reaction.add_metabolites(met_to_add)
                reactions_to_change.append(reaction)

        return reactions_to_change




    def __swap_bigg_metabolite(self, model_metabolite, new_compound_ontology_id):
        """
        Method to swap a metabolite in the BiGG database format.
        This method tries to find the BiGG format of the new compound. If it does not find, it will change into the
        Model SEED format.

        :param Model.Metabolite model_metabolite: metabolite to be replaced.
        :param int new_compound_ontology_id: ontology identifier of the metabolite that will replace the other
        :return:
        """

        model_metabolite.annotation = {}
        compound_container = self.__compounds_ontology.get_node_by_ont_id(new_compound_ontology_id)
        model_seed_id = compound_container.model_seed_id


        if model_seed_id:
            bigg_ids = None
            aliases = self.__compoundsIdConverter.get_modelSeedIdToDb().get(model_seed_id)
            if "BiGG" in aliases:
                bigg_ids = self.__compoundsIdConverter.convert_modelSeedId_into_other_dbID(model_seed_id, "BiGG")
            elif "BiGG1" in aliases:
                bigg_ids = self.__compoundsIdConverter.convert_modelSeedId_into_other_dbID(model_seed_id, "BiGG1")

            if bigg_ids:

                bigg_metabolite  = None
                bigg_id = None
                found = False
                i = 0
                while not found and i < len(bigg_ids):
                    try:
                        bigg_id = bigg_ids[i]
                        bigg_metabolite = bigg.get_bigg_metabolite(bigg_id)
                        found = True
                    except:
                        i += 1
                if bigg_metabolite:

                    old_inchikey = ""
                    if "inchi_key" in  model_metabolite.annotation.keys():
                        old_inchikey = model_metabolite.annotation["inchi_key"]

                    old_aliases = AnnotationUtils.get_annotation_from_cobra_annotation(model_metabolite)
                    old_id = model_metabolite.id

                    new_aliases = self.__compoundsIdConverter.get_all_aliases_by_modelSeedID(compound_container.model_seed_id)
                    annotation = AnnotationUtils.get_compound_annotation_format_by_aliases(new_aliases)
                    model_metabolite.annotation = annotation

                    new_inchikey = ""
                    if not compound_container.generic:
                        new_inchikey = compound_container.inchikey
                        model_metabolite.annotation["inchi_key"] = compound_container.inchikey


                    model_metabolite.annotation["smiles"] = compound_container.smiles
                    model_metabolite.formula = bigg_metabolite.get("formulae")[0]

                    charges = bigg_metabolite.get("charges")
                    if charges: model_metabolite.charge = charges[0]

                    compartment = model_metabolite.compartment

                    if compartment is None:
                        model_metabolite.compartment="c"

                    model_metabolite.name = bigg_metabolite.get("name")

                    self.__check_if_id_is_used_in_model_and_delete_it(bigg_id + "_" + model_metabolite.compartment)


                    self.report_material[model_metabolite.id] = bigg_id + "_" + model_metabolite.compartment
                    model_metabolite.id = bigg_id + "_" + model_metabolite.compartment

                    self.mapper.update_maps(old_inchikey,new_inchikey,old_id,
                                       model_metabolite.id,new_compound_ontology_id,
                                       old_aliases,new_aliases)

                else:
                    self.__change_boimmg_format_metabolite(model_metabolite, compound_container)

            else:
                self.__change_boimmg_format_metabolite(model_metabolite, compound_container)

        else:
            self.__change_boimmg_format_metabolite(model_metabolite, compound_container)

    def __swap_kegg_metabolite(self, model_metabolite, new_compound_ontology_id):
        """
        Method to swap a metabolite in the KEGG database format.
        This method tries to find the KEGG format of the new compound. If it does not find, it will change into the
        Model SEED format.

        :param Model.Metabolite model_metabolite: metabolite to be replaced.
        :param int new_compound_ontology_id: ontology identifier of the metabolite that will replace the other
        :return:
        """

        old_inchikey = ""
        if "inchi_key" in model_metabolite.annotation.keys():
            old_inchikey = model_metabolite.annotation["inchi_key"]

        old_aliases = AnnotationUtils.get_annotation_from_cobra_annotation(model_metabolite)
        old_id = model_metabolite.id

        model_metabolite.annotation = {}
        compound_container = self.__compounds_ontology.get_node_by_ont_id(new_compound_ontology_id)
        model_seed_id = compound_container.model_seed_id
        new_aliases = self.__compoundsIdConverter.get_all_aliases_by_modelSeedID(model_seed_id)


        if model_seed_id in self.__compoundsIdConverter.get_modelSeedIdToDb().keys():

            if "KEGG" in self.__compoundsIdConverter.get_modelSeedIdToDb().get(model_seed_id):

                kegg_id = self.__compoundsIdConverter.convert_modelSeedId_into_other_dbID(model_seed_id, "KEGG")[0]
                kegg_metabolite = KeggCompound(kegg_id)


                annotation = AnnotationUtils.get_compound_annotation_format_by_aliases(new_aliases)
                model_metabolite.annotation = annotation

                new_inchikey = ""
                if not compound_container.generic:
                    model_metabolite.annotation["inchi_key"] = compound_container.inchikey
                    new_inchikey = compound_container.inchikey

                model_metabolite.annotation["smiles"] = compound_container.smiles
                model_metabolite.formula = kegg_metabolite.get_formula()
                model_metabolite.name = kegg_metabolite.get_name()
                model_metabolite.charge = 0
                compartment = model_metabolite.compartment

                self.__check_if_id_is_used_in_model_and_delete_it(kegg_id + "_" + compartment)

                self.report_material[model_metabolite.id] = kegg_id + "_" + compartment

                model_metabolite.id = kegg_id + "_" + compartment

                self.mapper.update_maps(old_inchikey, new_inchikey, old_id,
                                   model_metabolite.id, compound_container.id,
                                   old_aliases, new_aliases)

            else:
                self.__change_boimmg_format_metabolite(model_metabolite, compound_container)

        else:
            self.__change_boimmg_format_metabolite(model_metabolite, compound_container)

    def __change_boimmg_format_metabolite(self, model_metabolite: Metabolite, compound_container):
        """
        This method changes the id of the metabolite arbitrarily. The format of the metabolite is set for the
        ontology format.
        It is worth noting that the :param compound_container can be either a ModelSeedCompound or an OntologyCompound

        :param Model.Metabolite model_metabolite: metabolite to be replaced.
        :param CompoundContainer compound_container: compound that will replace the other.
        :return:
        """

        old_inchikey = ""
        if "inchi_key" in model_metabolite.annotation.keys():
            old_inchikey = model_metabolite.annotation["inchi_key"]

        old_aliases = AnnotationUtils.get_annotation_from_cobra_annotation(model_metabolite)
        old_id = model_metabolite.id

        aliases = self.__compoundsIdConverter.get_all_aliases_by_modelSeedID(compound_container.model_seed_id)
        annotation = {}
        if aliases:
            annotation = AnnotationUtils.get_compound_annotation_format_by_aliases(aliases)

        annotation["boimmg.compound"] = self.__compoundsAnnotationConfigs["BOIMMG_ID_CONSTRUCTION"]+\
                                        str(compound_container.id)

        new_aliases = {"BOIMMG":[annotation["boimmg.compound"]]}
        model_metabolite.annotation = annotation

        new_inchikey = ""
        if not compound_container.generic:
            model_metabolite.annotation["inchi_key"] = compound_container.inchikey
            new_inchikey = compound_container.inchikey


        model_metabolite.annotation["smiles"] = compound_container.smiles
        model_metabolite.formula = compound_container.formula
        model_metabolite.charge = compound_container.charge
        model_metabolite.name = compound_container.name
        compartment = model_metabolite.compartment

        id = self.__compoundsAnnotationConfigs["BOIMMG_ID_CONSTRUCTION"]+ str(compound_container.id)

        new_metabolite_id = id + "_" + compartment

        self.__check_if_id_is_used_in_model_and_delete_it(new_metabolite_id)

        self.report_material[model_metabolite.id] = new_metabolite_id
        model_metabolite.id = new_metabolite_id

        self.mapper.update_maps(old_inchikey, new_inchikey, old_id,
                           model_metabolite.id, compound_container.id,
                           old_aliases, new_aliases)


    def __swap_modelseed_compound(self, model_metabolite: Metabolite, new_compound_ontology_id: int):

        """
        Method to swap a metabolite in the Model SEED database format.
        This method tries to find the Model SEED format of the new compound. If it does not find, it will change into an
        arbitrary id with the Model SEED or the Ontology format.

        :param Model.Metabolite model_metabolite: metabolite to be replaced.
        :param int new_compound_ontology_id: ontology identifier of the metabolite that will replace the other
        :return:
        """

        old_inchikey = ""
        if "inchi_key" in model_metabolite.annotation.keys():
            old_inchikey = model_metabolite.annotation["inchi_key"]

        old_aliases = AnnotationUtils.get_annotation_from_cobra_annotation(model_metabolite)
        old_id = model_metabolite.id


        model_metabolite.annotation = {}
        compound_container = self.__compounds_ontology.get_node_by_ont_id(new_compound_ontology_id)

        new_aliases = self.__compoundsIdConverter.get_all_aliases_by_modelSeedID(compound_container.model_seed_id)
        annotation = AnnotationUtils.get_compound_annotation_format_by_aliases(new_aliases)

        model_metabolite.annotation = annotation

        new_inchikey = ""
        if not compound_container.generic:
            model_metabolite.annotation["inchi_key"] = compound_container.inchikey
            new_inchikey = compound_container.inchikey

        model_metabolite.annotation["smiles"] = compound_container.smiles
        model_metabolite.formula = compound_container.formula
        model_metabolite.charge = compound_container.charge
        compartment = model_metabolite.compartment
        db_id = compound_container.model_seed_id
        model_metabolite.name = compound_container.name

        if db_id:

            self.__check_if_id_is_used_in_model_and_delete_it(compound_container.model_seed_id + "_" + compartment)

            self.report_material[model_metabolite.id] = compound_container.model_seed_id + "_" + compartment

            model_metabolite.id = compound_container.model_seed_id + "_" + compartment

            self.mapper.update_maps(old_inchikey, new_inchikey, old_id,
                               model_metabolite.id, new_compound_ontology_id,
                               old_aliases, new_aliases)
        else:
            self.__change_boimmg_format_metabolite(model_metabolite, compound_container)

    def __check_if_id_is_used_in_model_and_delete_it(self, id:int ):
        """

        This method checks whether a compound ID is used in the model and if so, it is removed

        :param id: compound identifier
        :return:
        """

        inModel = self.model.metabolites.has_id(id)

        if inModel:
            metabolite = self.model.metabolites.get_by_id(id)
            self.model.remove_metabolites([metabolite])


    def __get_game_changer_conjugates(self):
        """
        This method searches for the conjugates of a given compound in model and assigns their conjugated acid or base;

        :return dict {ontology id : (metabolite already in model ? (boolean), COBRApy metabolite or ontology id)}
        """

        conjugated_acid_in_model = self.__compounds_ontology.get_predecessors_by_ont_id_rel_type(
            self.__metabolite_in_model,
            "conjugated_acid_of"
        )
        conjugated_base_in_model = self.__compounds_ontology.get_predecessors_by_ont_id_rel_type(
            self.__metabolite_in_model,
            "conjugated_base_of"
        )

        game_changer = {}

        if conjugated_base_in_model:

            conjugated_base_to_change = self.__compounds_ontology.get_predecessors_by_ont_id_rel_type(
                self.__new_metabolite,
                "conjugated_base_of"
            )
            game_changer.update(
                self.__add_conjugate_to_game_changer("base", conjugated_base_in_model, conjugated_base_to_change))

        if conjugated_acid_in_model:
            conjugated_acid_to_change = self.__compounds_ontology.get_predecessors_by_ont_id_rel_type(
                self.__new_metabolite,
                "conjugated_acid_of"
            )
            game_changer.update(
                self.__add_conjugate_to_game_changer("acid", conjugated_acid_in_model, conjugated_acid_to_change))

        return game_changer


    def __get_game_changer(self):

        """
        This method set the game changer. The game changer will be a dictionary with all of the necessary
        information to swap the compounds. The game changer will have a given compound intermediates, conjugated acid
        or/and base and the respective metabolites to replace them.

        :return dict: {ontology id : (metabolite already in model ? (boolean), COBRApy metabolite or ontology id)}
        """

        intermediates_in_model = self.__compounds_ontology.get_all_predecessors_by_ont_id_rel_type(
            self.__metabolite_in_model,
            "precursor_of")

        conjugated_acid_in_model = self.__compounds_ontology.get_predecessors_by_ont_id_rel_type(
            self.__metabolite_in_model,
            "conjugated_acid_of"
        )
        conjugated_base_in_model = self.__compounds_ontology.get_predecessors_by_ont_id_rel_type(
            self.__metabolite_in_model,
            "conjugated_base_of"
        )

        intermediates_to_change = self.__compounds_ontology.get_all_predecessors_by_ont_id_rel_type(
            self.__new_metabolite, "precursor_of")

        #  values   ----->  (met in model?, if true: cobraMetabolite else: ontId)
        game_changer = self.__add_to_game_changer(self.__metabolite_in_model, self.__new_metabolite)

        game_changer.update(self.__get_intermediates_game_changer(intermediates_in_model, intermediates_to_change))

        if conjugated_acid_in_model:
            conjugated_acid_to_change = self.__compounds_ontology.get_predecessors_by_ont_id_rel_type(
                self.__new_metabolite,
                "conjugated_acid_of"
            )
            game_changer.update(
                self.__add_conjugate_to_game_changer("acid", conjugated_acid_in_model, conjugated_acid_to_change))

            intermediates_in_model = self.__compounds_ontology.get_all_predecessors_by_ont_id_rel_type(
                conjugated_acid_in_model[0],
                "precursor_of")

            intermediates_to_change = self.__compounds_ontology.get_all_predecessors_by_ont_id_rel_type(
                conjugated_acid_to_change[0],
                "precursor_of")

            game_changer.update(self.__get_intermediates_game_changer(intermediates_in_model, intermediates_to_change))

        if conjugated_base_in_model:
            conjugated_base_to_change = self.__compounds_ontology.get_predecessors_by_ont_id_rel_type(
                self.__new_metabolite,
                "conjugated_base_of"
            )
            game_changer.update(
                self.__add_conjugate_to_game_changer("base", conjugated_base_in_model, conjugated_base_to_change))

            intermediates_in_model = self.__compounds_ontology.get_all_predecessors_by_ont_id_rel_type(
                conjugated_base_in_model[0],
                "precursor_of")

            intermediates_to_change = self.__compounds_ontology.get_all_predecessors_by_ont_id_rel_type(
                conjugated_base_to_change[0],
                "precursor_of")

            game_changer.update(self.__get_intermediates_game_changer(intermediates_in_model, intermediates_to_change))

        return game_changer

    def __add_conjugate_to_game_changer(self, type: str, conjugated_in_model: list, new_conjugated: list) -> dict:
        """
        This method add a conjugated acid or base to the game changer

        :param string type: "acid" or "base"
        :param list conjugated_in_model: ontology id of the conjugate to be replaced
        :param list new_conjugated: ontology id of the conjugate that will replace

        :return dict: {ontology id : (metabolite already in model ? (boolean), COBRApy metabolite or ontology id)}
        """

        game_changer = {}

        if type == "base":

            metabolite_added = self.__add_to_game_changer(conjugated_in_model[0], new_conjugated[0])
            game_changer.update(metabolite_added)

        elif type == "acid":

            metabolite_added = self.__add_to_game_changer(conjugated_in_model[0], new_conjugated[0])
            game_changer.update(metabolite_added)

        return game_changer

    def __add_to_game_changer(self, met_in_model, new_met):
        """
        This method add metabolites to the game changer

        :param int met_in_model: ontology id of the metabolite to be replaced
        :param int new_met: ontology id of the metabolite to be replaced

        :return dict: {ontology id : (metabolite already in model ? (boolean), COBRApy metabolite or ontology id)}
        """
        game_changer = {}
        metabolite_container = self.__compounds_ontology.get_node_by_ont_id(new_met)
        inchikey = metabolite_container.inchikey
        aliases = self.__compoundsIdConverter.get_all_aliases_by_modelSeedID(metabolite_container.model_seed_id)
        metabolites_check_in_model = self.mapper.check_metabolites_in_model(inchikey,aliases)

        if met_in_model not in self.__metabolites_not_to_change:
            if metabolites_check_in_model:
                game_changer[met_in_model] = (True,[])
                for metabolite in metabolites_check_in_model:

                    game_changer[met_in_model][1].append(metabolite) # (met in model?, if true: [cobraMetabolite] else: ontId)
            else:
                game_changer[met_in_model] = (False, new_met)

            return game_changer
        else:
            if not metabolites_check_in_model:

                if metabolite_container.model_seed_id:

                    ms_container = self.__compounds_db.get_compound_by_id(metabolite_container.model_seed_id)
                    metabolites = self.generate_new_metabolites(ms_container)

                else:

                    metabolites = self.generate_new_boimmg_metabolites(metabolite_container)


                self.model.add_metabolites(metabolites)


            return game_changer

    def _check_if_compound_is_generic(self, compound_id):
        """
        This method check if some compound is generic or not

        :param int compound_id: ontology id of the metabolite

        :return boolean: is generic or not
        """

        compound_container = self.__compounds_ontology.get_node_by_ont_id(compound_id)
        return compound_container.generic

    def __get_intermediates_game_changer(self, intermediates_in_model, intermediates_to_change):
        """
        This method finds the correspondent substitute of each intermediate in the model

        :param list intermediates_in_model: list of ontology ids of the intermediates of the metabolite to be changed
        :param list intermediates_to_change: list of ontology ids of the intermediates of the metabolite that will replace

        :return dict: {ontology id : (metabolite already in model? (boolean), COBRApy metabolite or ontology id) }
        """

        game_changer = {}
        if self.__type == 1:
            intermediates = self.__set_intermediates_in_game_changer_type1(intermediates_in_model,
                                                                           intermediates_to_change)
            game_changer.update(intermediates)

        elif self.__type == 2:
            intermediates = self.__set_intermediates_in_game_changer_type2(intermediates_in_model,
                                                                           intermediates_to_change)
            game_changer.update(intermediates)

        return game_changer

    def __set_intermediates_in_game_changer_type1(self, intermediates_in_model, intermediates_to_change):
        """
        This method finds the correspondent substitute of each intermediate in the model for the type 1 game change

        :param list intermediates_in_model: list of ontology ids of the intermediates of the metabolite to be changed
        :param list intermediates_to_change: list of ontology ids of the intermediates of the metabolite that will replace

        :return dict: {ontology id : (metabolite already in model? (boolean), COBRApy metabolite or ontology id) }
        """

        game_changer = {}
        for met in intermediates_in_model:
            parent_met_in_model = self.__compounds_ontology.get_successors_by_ont_id_rel_type(met, "is_a")
            i = 0
            found = False
            while not found and i < len(intermediates_to_change):
                parent2 = self.__compounds_ontology.get_successors_by_ont_id_rel_type(intermediates_to_change[i],
                                                                                        "is_a")
                if parent_met_in_model[0] == parent2[0]:
                    found = True

                    metabolite_added = self.__add_to_game_changer(met, intermediates_to_change[i])
                    game_changer.update(metabolite_added)

                i += 1
        return game_changer

    def __set_intermediates_in_game_changer_type2(self, intermediates_in_model, intermediates_to_change):
        """
        This method finds the correspondent substitute of each intermediate in the model for the type 2 game change

        :param list intermediates_in_model: list of ontology ids of the intermediates of the metabolite to be changed
        :param list intermediates_to_change: list of ontology ids of the intermediates of the metabolite that will replace

        :return dict: {ontology id : (metabolite already in model? (boolean), COBRApy metabolite or ontology id) }
        """

        game_changer = {}
        isGenericInModel = self._check_if_compound_is_generic(intermediates_in_model[0])
        self.reactions_swapper.set_type2_is_generic_in_model(isGenericInModel)

        if isGenericInModel:
            for met in intermediates_in_model:
                i = 0
                found = False
                while not found and i < len(intermediates_to_change):
                    parent_met_to_change = self.__compounds_ontology.get_successors_by_ont_id_rel_type(
                        intermediates_to_change[i], "is_a")

                    if met in parent_met_to_change:
                        game_changer.update(self.__add_to_game_changer(met, intermediates_to_change[i]))
                        found = True
                    i += 1
        else:
            for met in intermediates_in_model:
                i = 0
                found = False
                while not found and i < len(intermediates_to_change):
                    children = self.__compounds_ontology.get_predecessors_by_ont_id_rel_type(
                        intermediates_to_change[i], "is_a")

                    if met in children:
                        game_changer.update(self.__add_to_game_changer(met, intermediates_to_change[i]))
                        found = True
                    i += 1
        return game_changer


    def __get_metabolite_container_by_id(self, id):
        """
        This method searches for a given metabolite in the Model

        :param string id: id in the Model

        :return Model.metabolite: if the metabolite is found returns the metabolite, otherwise returns None
        """

        for met in self.model.metabolites:
            if met.id == id:
                return met
        return None



    def change_reaction_format(self,reaction):
        """
        This method changes a given reaction format.

        :param Model.Reaction reaction: reaction to be changed
        :return:
        """

        self.reactions_swapper.swap_reactions([reaction])


    def __check_if_compound_exists_in_model_by_ontology_id(self,ontology_id):

        """
        Method to check whether a compound exists in the model using BOIMMG ID

        :param int ontology_id: boimmg id
        :return list<Compound>: list of compounds in model with this ID (list of the same compound in different compartments)
        """

        container = self.__compounds_ontology.get_node_by_ont_id(ontology_id)
        inchikey = container.inchikey
        aliases={}
        if container.model_seed_id:
            aliases = self.__compoundsIdConverter.get_all_aliases_by_modelSeedID(container.model_seed_id)

        aliases.update(container.aliases)
        return self.mapper.check_metabolites_in_model(inchikey,aliases,container)

    def generate_new_metabolites(self,compound_container: ModelSeedCompound):
        """
        This method will generate a new set of compounds (same metabolite in different compartments)

        :param ModelSeedCompound compound_container: internal wrappers of the ModelSEED compounds
        :return list<Metabolite>: list of newly generated compounds
        """

        model_metabolites = \
            model_utilities.generate_model_compounds_by_database_format(self.model,
                                                                        compound_container.getDbId(),
                                                                        self.__compoundsIdConverter,
                                                                        self.__compounds_db,
                                                                        self.__model_database)

        self.model.add_metabolites(model_metabolites)
        self.mapper.add_new_metabolites_to_maps(model_metabolites)
        return model_metabolites

    def generate_new_boimmg_metabolites(self, metabolite_container: CompoundNode):

        """
        This method will generate a new set of compounds (same metabolite in different compartments)

        :param CompoundNode metabolite_container: wrapper of BOIMMG compound
        :return list<Metabolite>: list of newly generated compounds
        """

        metabolites = model_utilities.generate_boimmg_metabolites(self.model,
                                                                  metabolite_container,
                                                                  self.__model_database,
                                                                  self.__compoundsIdConverter,
                                                                  self.__compoundsAnnotationConfigs)

        self.model.add_metabolites(metabolites)

        self.mapper.add_new_metabolites_to_maps(metabolites)

        return metabolites


    def setType2_isGenericInModel(self,isGeneric):
        self.reactions_swapper.set_type2_is_generic_in_model(isGeneric)

    def set_model_mapper(self, mapper):
        self.mapper = mapper


import os
import sys

import cobra
from rdkit import RDLogger

from boimmgpy.definitions import ROOT_DIR
from boimmgpy.service.representation_problem_solvers.representation_redundant_case_solver import RedundantCaseSolver

DATABASE_CONF = ROOT_DIR + "/configs/database_settings.conf"

def test_case_study_rrc(uri,user,password):

    with open(DATABASE_CONF,"w") as file:
        file.write("uri="+str(uri)+"\n")
        file.write("user="+str(user)+"\n")
        file.write("password="+str(password))

    try:
        RDLogger.DisableLog('rdApp.*')

        model = cobra.io.read_sbml_model(ROOT_DIR + "/case_studies/redundant_representation_case/iJR904_mapped.xml")
        components = ["cpd00214", "cpd03847", "cpd05274", "cpd25615", "cpd05237"]

        solver = RedundantCaseSolver(model, "BiGG")

        solver.swap_from_generic(["cpd22513", "cpd15649"], components, True)
        # solver.generateISAreactions()

        os.remove(DATABASE_CONF)

    except:
        os.remove(DATABASE_CONF)
        raise Exception("Not well run")





if __name__=="__main__":
    import os
    os.getcwd()
    uri=sys.argv[1]
    user = sys.argv[2]
    password = sys.argv[3]
    test_case_study_rrc(uri, user, password)



import logging
import re
import time
from logging.handlers import TimedRotatingFileHandler

import cobra
from biocyc import biocyc
from boimmgpy import definitions
from cobra import Model

from boimmgpy.database.containers.compound_node import CompoundNode
from boimmgpy.database.accessors.compounds_database_accessor import CompoundsDBAccessor
from boimmgpy.database.databases_babel import AliasesTransformer
from boimmgpy.service.model_mapper import ModelMapper
from boimmgpy.service.revisor.compounds_revisor import CompoundsRevisor
from boimmgpy.service.network_modifiers.metabolite_swapper import MetaboliteSwapper
from boimmgpy.model_seed.model_seed_compound import ModelSeedCompound
from boimmgpy.utilities import model_utilities
from boimmgpy.id_converters.compounds_id_converter import CompoundsIDConverter
from boimmgpy.model_seed.model_seed_compounds_database import ModelSeedCompoundsDBRest, \
    ModelSeedCompoundsDBRaw
from boimmgpy.definitions import TOOL_CONFIG_PATH, ROOT_DIR, COMPOUNDS_ANNOTATION_CONFIGS_PATH, REACTIONS_ANNOTATION_CONFIGS_PATH
from boimmgpy.utilities import file_utilities
from boimmgpy.utilities.rest_access_utils import RestUtils

logPath = ROOT_DIR + "/logs"

# format the log entries, backup limit and others
formatter = logging.Formatter('%(asctime)s %(name)s %(levelname)s %(message)s')
handler = TimedRotatingFileHandler(logPath, when='midnight', backupCount=20)
handler.setFormatter(formatter)
logger = logging.getLogger(__name__)
logger.addHandler(handler)
logger.setLevel(logging.DEBUG)

PROGRESS_BAR = ROOT_DIR + "/service/logs/progress_bar.txt"


# PROGRESS_BAR = "/workdir/resultsWorker/progress_bar.txt"

class SimpleCaseSolver():

    def __init__(self, model: Model, database_format: str, db_accessor = CompoundsDBAccessor()):
        """
        Class constructor

        :param Model model: model being analysed
        :param database_format: database format (ModelSEED, BiGG or KEGG)
        :param modelseed_compoundsdb:
        """

        self.report_material = {}
        self.__universal_model = Model("universal_model")
        self.model = model
        self.__database_format = database_format
        self.__configs = file_utilities.read_conf_file(TOOL_CONFIG_PATH)
        self.__home_path__ = ROOT_DIR
        self.__compounds_ontology = db_accessor

        if isinstance(db_accessor,CompoundsDBAccessor):
            self.__modelseedCompoundsDb = ModelSeedCompoundsDBRaw()

        else:
            self.__modelseedCompoundsDb = ModelSeedCompoundsDBRest()

        self.__define_instance_variables()

    def dump_maps(self,destination:str):
        """
        It creates a dump of all the model metabolite maps

        :param str destination: destination folder
        :return:
        """

        self.__mapper.create_map_dump(destination)

    def load_maps(self,folder):
        """
        Load all the metabolite maps in a given folder

        :param str folder: folder path
        :return:
        """

        self.__mapper.upload_maps(folder)

    def map_model(self):
        """
        Creates maps for all the metabolites in the model

        :return:
        """

        start = time.time()
        self.write_in_progress_bar("mapping the model... ", 10)
        logger.info("mapping model")

        if isinstance(self.__compounds_ontology, CompoundsDBAccessor):
            self.__mapper.map_model(self.__database_format)

        else:
            cobra.io.write_sbml_model(self.__model, definitions.ROOT_DIR + "/temp/model_to_be_submitted.xml")
            response = RestUtils.map_model(definitions.ROOT_DIR + "/temp/model_to_be_submitted.xml", self.__database_format)
            maps = response.json().get("result")

            self.__mapper.boimmg_db_model_map = maps["boimmg_db_model_map"]
            self.__mapper.boimmg_db_model_map_reverse = {int(k): maps["boimmg_db_model_map_reverse"][k] for k, v in maps["boimmg_db_model_map_reverse"].items()}
            self.__mapper.compounds_aliases_indexation = maps["compounds_aliases_indexation"]
            self.__mapper.compounds_aliases_indexation_reverse = maps["compounds_aliases_indexation_reverse"]
            self.__mapper.compound_inchikey_indexation = maps["compound_inchikey_indexation"]

        self.__mapper.mapped = True

        logger.info("model mapped")

        self.write_in_progress_bar("model mapped ", 31)
        finished = time.time()
        logger.info("mapping finished: %d" % (finished - start))

    def write_in_progress_bar(self, message: str, value: float):

        """
        Write information related to the progress bar

        :param str message: message of what is being done
        :param float value: value to assign in the progress bar
        :return:
        """

        with open(PROGRESS_BAR,"w") as file:
            file.write(message + ":" + str(value))

    def get_progress_bar_state(self) -> float:
        """
        This method reads the progress bar message and the respective value

        :return float state: value to assign in the progress bar
        """
        with open(PROGRESS_BAR,"r") as file:
            line = file.read()
            line = line.strip()
            state = line.split(":")[1]
            return float(state)

    def calculate_division_for_progress_bar(self,state,total_processes_left):
        left = 100 - state
        for_each_part = left / total_processes_left

        return for_each_part

    @property
    def model(self):
        return self.__model

    @model.setter
    def model(self, value):
        if isinstance(value, Model):
            self.__model = value
        else:
            raise ValueError("introduce a cobrapy model")


    def __define_instance_variables(self):
        """
        Method to define a set of instance values

        :return:
        """

        self.__swapped = []
        self.lineage = []


        self.__compoundsAnnotationConfigs = file_utilities.read_conf_file(
            COMPOUNDS_ANNOTATION_CONFIGS_PATH)

        self.__reactionsAnnotationConfigs = file_utilities.read_conf_file(
            REACTIONS_ANNOTATION_CONFIGS_PATH)

        self.__compoundsAnnotationConfigs = file_utilities.read_conf_file(COMPOUNDS_ANNOTATION_CONFIGS_PATH)
        self.__reactionsAnnotationConfigs = file_utilities.read_conf_file(REACTIONS_ANNOTATION_CONFIGS_PATH)
        self.__compoundsIdConverter = CompoundsIDConverter()


        self.__not_to_change_classes = ["cpd03476"]
        self.__set_not_to_change_compounds()

        self.__compounds_revisor = CompoundsRevisor(self.model, self.__universal_model)
        biocyc.set_organism("meta")

        self.__swapper = MetaboliteSwapper(self.model, None, None, 0,
                                           self.__modelseedCompoundsDb,
                                           self.__compounds_ontology,
                                           model_database=self.__database_format,
                                           compoundsIdConverter=self.__compoundsIdConverter,
                                           universal_model=self.__universal_model,
                                           not_to_change_compounds=self.__not_to_change_compounds
                                           )

        self.__mapper = ModelMapper(self.model,
                                    self.__compoundsIdConverter,self.__compounds_ontology)

        self.__swapper.set_model_mapper(self.__mapper)
        self.__compounds_revisor.set_model_mapper(self.__mapper)

    def __set_not_to_change_compounds(self):
        """
        One would want to protect few compounds of being changed.

        :return:
        """
        self.__not_to_change_compounds = []
        for compound_class in self.__not_to_change_classes:
            parent_id = self.__compounds_ontology.get_node_id_from_model_seed_id(compound_class)
            children = self.__compounds_ontology.get_predecessors_by_ont_id_rel_type(parent_id, "is_a")
            self.__not_to_change_compounds.extend(children)


    def swap_compound(self, compound_in_model:str, compound_to_change: str, compounds_left=1):
        """
        This method aims at swapping all the biosynthetic precursors, and the conjugated acid and base of a given target.
        It accepts all types of changes.

        :param str compound_in_model: compound in model
        :param str compound_to_change: compound to replace the one in the model
        :param str compounds_left: compounds left in the swapping operation
        :return:
        """

        if not self.__mapper.mapped:

            self.map_model()

        compound_in_model = self.transform_input_into_boimmg_ids(compound_in_model)
        if not compound_to_change:
            exit(2)

        compound_to_change = self.transform_input_into_boimmg_ids(compound_to_change)
        if not compound_to_change:
            exit(2)

        compound_in_model_node = self.__compounds_ontology.get_node_by_ont_id(compound_in_model)
        compound_to_change_node = self.__compounds_ontology.get_node_by_ont_id(compound_to_change)

        state = self.get_progress_bar_state()
        per_iteration = self.calculate_division_for_progress_bar(state,compounds_left)


        self.write_in_progress_bar("Swapping "+compound_in_model_node.name
                                   + " to " + compound_to_change_node.name,state)

        logger.info("Swapping "+compound_in_model_node.name + " to " + compound_to_change_node.name)

        type = 0

        if compound_in_model_node.generic and not compound_to_change_node.generic:
            parents = self.__compounds_ontology.get_successors_by_ont_id_rel_type(compound_to_change, "is_a")
            if compound_in_model in parents:
                type = 2
            else:
                type = 3

        elif compound_to_change_node.generic and not compound_in_model_node.generic:
            parents = self.__compounds_ontology.get_successors_by_ont_id_rel_type(compound_in_model, "is_a")
            if compound_to_change in parents:
                type = 2
            else:
                type = 3

        elif not compound_to_change_node.generic and not compound_in_model_node.generic:

            parents1 = self.__compounds_ontology.get_successors_by_ont_id_rel_type(compound_in_model, "is_a")
            parents2 = self.__compounds_ontology.get_successors_by_ont_id_rel_type(compound_to_change, "is_a")
            if parents1 and parents2:
                if parents1[0] == parents2[0]:
                    type = 1
                else:
                    type = 2

        if type != 0:
            self.__swapper.set_type(type)
            self.__swapper.set_compound_in_model(compound_in_model)
            self.__swapper.set_new_compound(compound_to_change)

            self.__swapper.swap_metabolites()

        else:
            self.__swapper.set_type(type)
            self.__swapper.set_compound_in_model(compound_in_model)
            self.__swapper.set_new_compound(compound_to_change)

            self.__swapper.swap_only_metabolites_and_conjugates()

        self.write_in_progress_bar("Swap performed", per_iteration)
        logger.info("Swap performed")


    def write_report(self,file_path):

        metabolites_report_material = self.__swapper.report_material
        reactions_report_material = self.__swapper.reactions_swapper.report_material

        with open(file_path,"w") as f:

            f.write("--Metabolites--\n")
            f.write("metabolite in model,replacer metabolite\n")

            for metabolite in metabolites_report_material:
                f.write(metabolite+",")
                f.write(metabolites_report_material[metabolite]+"\n")

            f.write("--Reactions--\n")
            f.write("reaction in model,replacer reaction\n")

            for reaction in reactions_report_material:
                f.write(reaction+",")
                f.write(reactions_report_material[reaction]+"\n")


    def __check_if_compound_exists_in_model_by_ontology_id(self, ontology_id: int) -> list:
        """
        This method will check whether a BOIMMG compound exists in the model.

        :param int ontology_id: BOIMMG id
        :return list: compounds in model
        """
        res = []

        for model_id, boimmg_id in self.__mapper.boimmg_db_model_map.items():
            if boimmg_id == ontology_id:
                model_compound_container = self.model.metabolites.get_by_id(model_id)
                res.append(model_compound_container)

        return res


    def __replace_electron_donors_and_acceptors(self):
        """
        This method will swap all the generic electron donors and acceptors eventually added when introducing
        MetaCyc reactions.

        :return:
        """

        donor_model_seed_id = "cpd26978"
        electron_transfer_quinol = 749301

        ontology_id = self.__compounds_ontology.get_node_id_from_model_seed_id(donor_model_seed_id)
        container = self.__compounds_ontology.get_node_by_ont_id(ontology_id)
        inchikey = container.inchikey
        aliases = self.__compoundsIdConverter.get_all_aliases_by_modelSeedID(donor_model_seed_id)
        metabolites_in_model = self.__mapper.check_metabolites_in_model(inchikey, aliases)
        self.__swapper.set_compound_in_model(ontology_id)
        self.__swapper.set_type(0)

        if metabolites_in_model:

            leaves = self.__compounds_ontology.get_leaves_from_ont_id(ontology_id)

            for metabolite in metabolites_in_model:
                compartment = metabolite.compartment
                found_leaf = False
                i = 0
                while not found_leaf and i < len(leaves):
                    parents = self.__compounds_ontology.get_all_parents(leaves[i])
                    if electron_transfer_quinol not in parents:

                        leaf_container = self.__compounds_ontology.get_node_by_ont_id(leaves[i])
                        modelseedid = leaf_container.model_seed_id
                        inchikey = leaf_container.inchikey
                        aliases = self.__compoundsIdConverter.get_all_aliases_by_modelSeedID(modelseedid)

                        metabolites_in_model_to_change = \
                            self.__mapper.check_metabolites_in_model(inchikey, aliases,
                                                                     leaf_container)

                        found_compartment_leaf = False
                        j = 0
                        while not found_compartment_leaf and j < len(metabolites_in_model_to_change):
                            if metabolites_in_model_to_change[j].compartment == compartment:
                                found_compartment_leaf = True
                                found_leaf = True

                                self.__swapper.set_new_compound(leaves[i])
                                self.__swapper.swap_only_metabolites_and_conjugates(metabolite,
                                                                                    metabolites_in_model_to_change[j])
                            else:
                                j += 1
                    i += 1



    def __swap_metabolites(self, metabolite_in_model: int, new_metabolite: int, type: int, target=False) -> list:
        """
        This method aims at calling the MetaboliteSwapper to swap a given metabolite, their biosynthetic precursors
        and their conjugated base and acid, eventually.

        :param int metabolite_in_model: BOIMMG id of the metabolite in model
        :param int new_metabolite: BOIMMG id of the new metabolite
        :param int type: type of swap
        :param boolean target: whether the new metabolite is the target metabolite or not
        :return:
        """

        self.__swapper.set_compound_in_model(metabolite_in_model)
        self.__swapper.set_new_compound(new_metabolite)
        self.__swapper.set_type(type)

        if type == 2:
            self.__type_2_isGeneric_setter(metabolite_in_model)

        if not target:
            changed_reactions = self.__swapper.swap_only_metabolites_and_conjugates()
            self.__swapped.append(new_metabolite)
        else:
            changed_reactions = self.__swapper.swap_metabolites()

        self.__universal_model = self.__swapper.get_universal_model()

        return changed_reactions

    def __type_2_isGeneric_setter(self, metabolite_in_model):

        container = self.__compounds_ontology.get_node_by_ont_id(metabolite_in_model)
        if container.generic:
            self.__swapper.setType2_isGenericInModel(True)
        else:
            self.__swapper.setType2_isGenericInModel(False)

    def __check_and_change_all_children(self, parent: int, new_metabolite: int, target=False) -> list:
        """
        This method absorbs all the compounds of a specific chemical type and swap them into the :param new_metabolite

        :param int parent: BOIMMG id of the structural parent
        :param int new_metabolite: BOIMMG id of the metabolite that will prevail
        :param boolean target: whether the new metabolite is the target or not
        :return list: changed reactions
        """

        children = self.__compounds_ontology.get_predecessors_by_ont_id_rel_type(parent, "is_a")
        changed_reactions = []

        for child in children:
            if child != new_metabolite:

                child_container = self.__compounds_ontology.get_node_by_ont_id(child)

                aliases = AliasesTransformer.transform_boimmg_aliases_into_model_seed(child_container.aliases)

                model_seed_id = child_container.model_seed_id

                if model_seed_id:

                    converted_aliases = self.__compoundsIdConverter.get_all_aliases_by_modelSeedID(model_seed_id)

                    merged_aliases = self.merge_dictionaries(converted_aliases, aliases)

                    child_in_model = self.__mapper.check_metabolites_in_model(child_container.inchikey,
                                                                              merged_aliases)

                else:

                    child_in_model = self.__mapper.check_metabolites_in_model(child_container.inchikey,
                                                                              child_container.aliases,
                                                                              child_container)

                if child_in_model:
                    changed_reactions = self.__swap_metabolites(child, new_metabolite, 1, target)

        return changed_reactions

    def convert_aliases_into_mapper_format(self,aliases):
        """
        Such method can convert the aliases from retrieved from ModelSEED into the mapper format

        :param dict aliases: aliases from ModelSEED
        :return dict new_aliases: the converted dictionary
        """
        new_aliases = {}
        for alias in aliases:
            if alias in self.__compoundsAnnotationConfigs.keys():
                new_key = self.__compoundsAnnotationConfigs[alias]
                new_aliases[new_key] = aliases[alias]

        return new_aliases

    @staticmethod
    def merge_dictionaries(dict1, dict2):
        for key in dict1:
            if key in dict2:
                dict1[key].extend(dict2[key])
        return dict1


    def generate_new_metabolite(self, compound_container):
        """
        Such method generates a whole new metabolite using a ModelSEED compound wrapper or a CompoundNode

        :param compound_container: ModelSEED compound wrapper
        :return list: list of newly generated metabolites
        """

        if isinstance(compound_container, ModelSeedCompound):
            model_metabolites = \
                model_utilities.generate_model_compounds_by_database_format(self.model,
                                                                           compound_container.getDbId(),
                                                                           self.__compoundsIdConverter,
                                                                           self.__modelseedCompoundsDb,
                                                                           self.__database_format)
        elif isinstance(compound_container, CompoundNode):
            model_metabolites = model_utilities.generate_model_compounds_by_database_format(self.model,
                                                                                           compound_container.model_seed_id,
                                                                                           self.__compoundsIdConverter,
                                                                                           self.__modelseedCompoundsDb,
                                                                                           self.__database_format)

        else:
            raise ValueError("Not expected type")

        self.__model.add_metabolites(model_metabolites)
        self.__mapper.add_new_metabolites_to_maps(model_metabolites)
        return model_metabolites

    def generate_new_boimmg_metabolite(self, boimmg_container: CompoundNode) -> list:

        """
        Such method generates new BOIMMG compounds and adds them to the model (same compound in different compartments)

        :param CompoundNode boimmg_container: BOIMMG container for model addition
        :return list<Metabolite>: list of model metabolites in different compartments
        """

        model_metabolites = model_utilities.generate_boimmg_metabolites(self.model,
                                                                       boimmg_container,
                                                                       self.__database_format,
                                                                       self.__compoundsIdConverter,
                                                                       self.__compoundsAnnotationConfigs,
                                                                       )

        self.__model.add_metabolites(model_metabolites)
        self.__mapper.add_new_metabolites_to_maps(model_metabolites)
        return model_metabolites

    def add_metabolites_to_model(self, metabolites):

        for metabolite in metabolites:

            container = self.__compounds_ontology.get_node_by_ont_id(metabolite)
            metabolites_in_model = self.__mapper.check_metabolites_in_model(container.inchikey, container.aliases,
                                                                            container)

            if not metabolites_in_model:

                if container.model_seed_id:
                    ms_container = self.__modelseedCompoundsDb.get_compound_by_id(container.model_seed_id)
                    self.generate_new_metabolite(ms_container)

                else:
                    self.generate_new_boimmg_metabolite(container)

    def transform_input_into_boimmg_ids(self, target: str) -> str:
        """
        Method that converts a target ID from an external database into BOIMMG's

        :param str target: target identifier of an external database
        :return str: boimmg identifier
        """

        accessor = CompoundsDBAccessor()

        boimmg_prefix = self.__compoundsAnnotationConfigs["BOIMMG_ID_CONSTRUCTION"]
        target_res = self._transform_database_ids_into_boimmg_ids(target,accessor,boimmg_prefix)


        return target_res

    def _transform_database_ids_into_boimmg_ids(self, target: str, accessor: CompoundsDBAccessor, boimmg_prefix: str)-> str:

        """
        Such method transforms a target ID from an external database into BOIMMG's

        :param str target: target identifier from an external database
        :param CompoundsDBAccessor accessor:
        :param str boimmg_prefix: BOIMMG prefix
        :return:
        """

        targets_res=''
        if boimmg_prefix in target:
            targets_res = int(target.replace(boimmg_prefix, ""))


        elif re.match("^[0-9]*$", target):

            targets_res = int(target)

        elif "cpd" not in target:
            model_seed_ids = self.__compoundsIdConverter.convert_db_id_to_model_seed_by_db_id(target)

        else:
            model_seed_ids = [target]

        if not targets_res:
            for model_seed_id in model_seed_ids:
                node1 = accessor.get_node_from_model_seed_id(model_seed_id)

                if node1:
                    targets_res = node1.id
                    break

        return targets_res


import copy
import itertools
import re

from rdkit import Chem, DataStructs
from rdkit.Chem import AllChem, MolToSmiles, MolFromSmiles, MolFromSmarts


def check_n_to_m_atoms(molecule,n,m,atom_number):
    atoms = molecule.GetAtoms()
    n_atoms = 0
    for atom in atoms:
        if atom.GetAtomicNum() == atom_number:
            n_atoms+=1

    if n_atoms>=n and n_atoms<=m:
        return True

    else:
        return False

def check_at_least_n_atoms(molecule,n,atom_number):
    atoms = molecule.GetAtoms()
    n_atoms = 0
    for atom in atoms:
        if atom.GetAtomicNum() == atom_number:
            n_atoms += 1

    if n_atoms >= n:
        return True

    else:
        return False

def check_if_odd_number_of_atom(molecule,atom_number,backbone_n_atoms=0):
    atoms = molecule.GetAtoms()
    n_atoms = backbone_n_atoms
    for atom in atoms:
        if atom.GetAtomicNum() == atom_number:
            n_atoms += 1

    if n_atoms%2 != 0:
        # print(MolToSmiles(molecule))
        return True

    else:
        return False

def get_side_chains(parent_smiles,child_smiles):

    r_groups_number = parent_smiles.count("*")

    neutralized_smiles, _ = NeutraliseCharges(parent_smiles)

    transformation1 = convert_model_seed_dummy_atoms(parent_smiles) + ">>"
    component_smiles = "*C(=O)(O)"

    component_smiles_list = []
    for i in range(1, r_groups_number + 1):
        component_smiles_list.append(component_smiles.replace("*", "[*:" + str(i) + "]"))

    transformation2 = convert_model_seed_dummy_atoms(neutralized_smiles) + ">>"

    transformation1 += ".".join(component_smiles_list)
    transformation2 += ".".join(component_smiles_list)

    transformation_smarts1 = AllChem.ReactionFromSmarts(transformation1)

    transformation_smarts2 = AllChem.ReactionFromSmarts(transformation2)

    child_mol = MolFromSmiles(child_smiles)

    try:
        new_molecules = [Chem.MolToSmiles(x, 1) for x in transformation_smarts1.RunReactants(
            tuple([child_mol]))[0]]

        return new_molecules

    except:
        try:
            new_molecules = [Chem.MolToSmiles(x, 1) for x in transformation_smarts2.RunReactants(
                tuple([child_mol]))[0]]

            return new_molecules

        except:
            return None

    return None




def check_if_only_cis_stereochemistry(molecule):

    smiles = MolToSmiles(molecule)

    if "=" in smiles:
        regex_cis = re.compile(r"/[A-Z]{1}[a-z]{0,1}=[A-Z]{1}[a-z]{0,1}\\")
        regex_trans = re.compile(r"/[A-Z]{1}[a-z]{0,1}=[A-Z]{1}[a-z]{0,1}/|\\[A-Z]{1}[a-z]{0,1}=[A-Z]{1}[a-z]{0,1}\\")
        cis_stereochemistry = regex_cis.findall(smiles)
        trans_stereochemistry = regex_trans.findall(smiles)
        double_bonds = smiles.count("=")

    else:
        return True

    if cis_stereochemistry and not trans_stereochemistry and len(cis_stereochemistry) == double_bonds:
        return True
    else:
        return  False
    

def get_atoms_idx_from_molecule(molecule):
    res = []
    atoms = molecule.GetAtoms()
    for atom in atoms:
        res.append(atom.GetIdx())
    return res

def checkIfCycle(molecule):
    info = molecule.GetRingInfo()
    numRings = info.NumRings()
    if numRings == 0:
        return True
    else:
        return False

def __check_if_sidechains_are_equal(lst1,lst2):
    lst_1 = copy.deepcopy(lst1)
    lst_2 = copy.deepcopy(lst2)
    for mol1 in lst_1:
        i = 0
        found = False
        while i < len(lst_2):
            fgp1 = AllChem.GetMorganFingerprint(mol1,1)
            fgp2 = AllChem.GetMorganFingerprint(lst2[i],1)
            similarity = DataStructs.TanimotoSimilarity(fgp1, fgp2)
            if similarity == 1:
                lst_2.remove(lst_2[i])
                found = True
            else:
                i += 1

        if not found:
            return False

    if not lst_2:
        return True

    else:
        return False

def check_if_sidechains_are_equal(lst1,lst2):

    if len(lst1) <= len(lst2):
        return __check_if_sidechains_are_equal(lst2,lst1)

    else:
        return __check_if_sidechains_are_equal(lst1, lst2)

def check_if_molecules_are_equal(mol1,mol2):
    fg1 = AllChem.GetMorganFingerprint(mol1,1,useFeatures=True,useChirality=True)
    fg2 = AllChem.GetMorganFingerprint(mol2,1,useFeatures=True,useChirality=True)
    similarity = DataStructs.TanimotoSimilarity(fg1, fg2)
    if similarity==1:
        return True
    else:
        return False

def retrieve_fragments(mol):
    substructures_smiles = MolToSmiles(mol)
    sidechains = []
    sidechains_smiles_list = substructures_smiles.split(".")
    for sidechain_smiles in sidechains_smiles_list:
        sidechain = MolFromSmiles(sidechain_smiles)
        if sidechain_smiles != "":
            sidechains.append(sidechain)
    return sidechains,sidechains_smiles_list

def convert_model_seed_dummy_atoms(modelseed_smiles):
    i=1
    new_smart=""
    match1 = re.finditer("\*",modelseed_smiles)

    previous = None
    for m in match1:
        start = m.start()
        end = m.end()
        if not previous:
            new_smart += modelseed_smiles[:start] + "[*:" + str(i) + "]"
        else:
            new_smart += modelseed_smiles[previous:start] + "[*:" + str(i) + "]"

        i+=1
        previous = end

    if previous!=len(modelseed_smiles)-1:
        new_smart+=modelseed_smiles[previous:]
    return new_smart

def _InitialiseNeutralisationReactions():
    patts= (
        # Imidazoles
        ('[n+;H]','n'),
        # Amines
        ('[N+;!H0]','N'),
        # Carboxylic acids and alcohols
        ('[$([O-]);!$([O-][#7])]','O'),
        # Thiols
        ('[S-;X1]','S'),
        # Sulfonamides
        ('[$([N-;X2]S(=O)=O)]','N'),
        # Enamines
        ('[$([N-;X2][C,N]=C)]','N'),
        # Tetrazoles
        ('[n-]','[nH]'),
        # Sulfoxides
        ('[$([S-]=O)]','S'),
        # Amides
        ('[$([N-]C=O)]','N'),
        )
    return [(Chem.MolFromSmarts(x),Chem.MolFromSmiles(y,False)) for x,y in patts]

_reactions=None

def NeutraliseCharges(smiles, reactions=None):
    global _reactions
    if reactions is None:
        if _reactions is None:
            _reactions=_InitialiseNeutralisationReactions()
        reactions=_reactions
    mol = Chem.MolFromSmiles(smiles)
    replaced = False
    for i,(reactant, product) in enumerate(reactions):
        while mol.HasSubstructMatch(reactant):
            replaced = True
            rms = AllChem.ReplaceSubstructs(mol, reactant, product)
            mol = rms[0]
    if replaced:
        return (Chem.MolToSmiles(mol,True), True)
    else:
        return (smiles, False)

def check_similarity_between_generic_and_complete_representation(generic_smiles,complete_smiles):
    complete_smiles,_ = NeutraliseCharges(complete_smiles)
    generic_smiles,_ = NeutraliseCharges(generic_smiles)
    complete_mol = MolFromSmiles(complete_smiles)
    generic_mol = MolFromSmarts(generic_smiles)
    match = complete_mol.GetSubstructMatch(generic_mol)
    if match:
        return True
    return False


def __check_if_dummy_is_undefined(start,strange_metabolite):
    for sm in strange_metabolite:
        if start == sm + 1:
            return True
    return False


def get_combinations_of_dummy_atoms(modelseed_smiles):

    r_groups = modelseed_smiles.count("*")
    to_subtract = modelseed_smiles.count("!*")
    r_groups = r_groups-to_subtract

    lst = [i for i in range(1,r_groups+1)]
    combinations = itertools.permutations(lst)
    combinations = list(combinations)

    res = []
    for comb in combinations:
        lst= list(comb)

        new_smart = ""
        match1 = re.finditer("\*", modelseed_smiles)
        strange_metabolite = re.finditer("!", modelseed_smiles)
        strange_metabolite_positions = [sm.start() for sm in strange_metabolite]

        i=0
        previous = None
        for m in match1:
            start = m.start()
            end = m.end()
            undefined_met = False
            if strange_metabolite:
                undefined_met =  __check_if_dummy_is_undefined(start,strange_metabolite_positions)

            if not undefined_met:
                if not previous:
                    new_smart += modelseed_smiles[:start] + "[*:" + str(lst[i]) + "]"
                else:
                    new_smart += modelseed_smiles[previous:start] + "[*:" + str(lst[i]) + "]"

                i += 1
            else:
                new_smart += modelseed_smiles[previous:start+1]
                new_smart = new_smart.replace("!","")


            previous = end



        if previous != len(modelseed_smiles) - 1:
            new_smart += modelseed_smiles[previous:]

        res.append(new_smart)

    return res

def check_if_molecule_has_double_bonds(molecule):
    atoms = molecule.GetAtoms()
    for atom in atoms:
        bonds = atom.GetBonds()
        for bond in bonds:
            if bond.GetBondType() == Chem.BondType.DOUBLE:
                return True
    return False

def check_if_atom_has_double_bonds(atom):
    bonds = atom.GetBonds()
    for bond in bonds:
        if bond.GetBondType() == Chem.BondType.DOUBLE:
            return True
    return False

def check_if_atom_has_triple_bonds(atom):
    bonds = atom.GetBonds()
    for bond in bonds:
        if bond.GetBondType() == Chem.BondType.TRIPLE:
            return True
    return False

def check_if_two_molecules_are_equal_from_smiles(smiles1,smiles2):
    mol1 = MolFromSmiles(smiles1)
    mol2 = MolFromSmiles(smiles2)

    fgp1 = AllChem.GetMorganFingerprint(mol1, 1,useFeatures =True,useChirality=True)
    fgp2 = AllChem.GetMorganFingerprint(mol2, 1,useFeatures =True,useChirality=True)
    similarity = DataStructs.TanimotoSimilarity(fgp1, fgp2)

    if similarity == 1:
        return True

    else:
        return False






import copy


class SplitMolecule(object):
    def __init__(self, complete_molecule, core, sidechains, ontid=None, dbid=None):
        self.__complete_molecule = complete_molecule
        self.__core__ = core
        self.__sidechains = sidechains
        self.__db_id = dbid
        self.__ontology_id__ = ontid

    def getSidechains(self):
        return self.__sidechains

    def getDbId(self):
        return self.__db_id

    def getOntID(self):
        return self.__ontology_id__

    def __copy__(self):
        return SplitMolecule(self.__complete_molecule, self.__core__,self.__sidechains,self.__ontology_id__,self.__db_id)

    def __deepcopy__(self, memo):
        return SplitMolecule(copy.deepcopy(self.__complete_molecule),
                             copy.deepcopy(self.__core__),copy.deepcopy(self.__sidechains),
                             copy.deepcopy(self.__ontology_id__),copy.deepcopy(self.__db_id))


import math
import re
from collections import defaultdict

from cobra import Model, Reaction
from sympy import solve, Symbol, Integer

from boimmgpy.database.accessors.compounds_database_accessor import CompoundsDBAccessor
from boimmgpy.utilities import file_utilities
from boimmgpy.utilities.annotation_utils import AnnotationUtils
from boimmgpy.definitions import COMPOUNDS_ANNOTATION_CONFIGS_PATH


class CompoundsRevisor:

    def __init__(self,model,universal_model = None,compoundsAnnotationConfigs = None):
        if not universal_model:
            self.__universal_model = Model()
        else:
            self.__universal_model=universal_model

        if not compoundsAnnotationConfigs:
            self.compoundsAnnotationConfigs = file_utilities.read_conf_file(COMPOUNDS_ANNOTATION_CONFIGS_PATH)

        else:
            self.compoundsAnnotationConfigs = compoundsAnnotationConfigs

        self.__model = model
        self.__get_hydrogen_from_model()
        self.__changed_reaction = None

    def __lp_to_balance_reaction(self,eq):

        Ls = list('abcdefghijklmnopqrstuvwxyz')

        Ss, Os, Es, a, i = defaultdict(list), Ls[:], [], 1, 1

        for p in eq.split('->'):
            for k in p.split('+'):
                c = [Ls.pop(0), 1]
                for e, m in re.findall('([A-Z][a-z]?)([0-9]*)', k):
                    m = 1 if m == '' else int(m)
                    a *= m
                    d = [c[0], c[1] * m * i]
                    Ss[e][:0], Es[:0] = [d], [[e, d]]
            i = -1


        Ys = dict((s, eval('Symbol("' + s + '")')) for s in Os if s not in Ls)

        Qs = [eval('+'.join('%d*%s' % (c[1], c[0]) for c in Ss[s]), {}, Ys) for s in Ss] + [Ys['a'] - a]

        k = solve(Qs, *Ys)

        # k = lcm(Qs, *Ys)
        new_k = k.copy()
        defined_variable_value = 0
        symbols_list = []
        list_of_equivalent_symbols = []
        for symbol in k:

            if isinstance(k[symbol], Symbol):
                new_k[k[symbol]] = symbol
                list_of_equivalent_symbols.append(symbol)

            t = k[symbol]
            present_symbols = t.free_symbols
            for s in present_symbols:
                if s not in symbols_list:
                    symbols_list.append(s)

            if isinstance(k[symbol], Integer):
                defined_variable_value = k[symbol]

        if len(symbols_list) == 1 and defined_variable_value != 0:

            to_process_symbol = symbols_list[0]
            for symbol in new_k:

                if not isinstance(new_k[symbol], Integer):
                    t = new_k[symbol]

                    new = t.subs({to_process_symbol: defined_variable_value})

                    new_k[symbol] = new

                if new_k[symbol] in list_of_equivalent_symbols:
                    t = new_k[symbol]

                    new = t.subs({new_k[symbol]: defined_variable_value})

                    new_k[symbol] = new

        return new_k, Ys

    def balance_reaction(self,reaction: Reaction):

        eq = self.__get_reaction_equation(reaction)

        try:
            k,Ys = self.__lp_to_balance_reaction(eq)
        except:
            return False

        isHydrogenInReaction = True

        if len(reaction.compartments) == 1:
            hydrogen = self.__hydrogens_in_model.get(list(reaction.compartments)[0])
            isHydrogenInReaction = self.__check_if_compound_in_reaction(reaction, hydrogen)

        if len(k) == len(Ys) and 0 not in k.values():

            try:
                stoich = self.__process_result_of_new_balanced_reaction(k,Ys,eq,reaction)
                reaction.add_metabolites(stoich,combine=False)
                return True

            except:
                return False
            # print('->'.join('+'.join(pM(N.pop(0)) + str(t) for t in p.split('+')) for p in eq.split('->')))

        elif not isHydrogenInReaction:

            reactants_products = eq.split("->")
            reactants_products[0] += "+H"

            new_eq = "->".join(reactants_products)
            k, Ys = self.__lp_to_balance_reaction(new_eq)

            if len(k) == len(Ys) and 0 not in k.values():

                new_reaction = reaction.copy()

                new_reaction.add_metabolites({hydrogen: -1})

                try:
                    stoich = self.__process_result_of_new_balanced_reaction(k, Ys, new_eq, new_reaction)

                    reaction.add_metabolites(stoich, combine=False)
                    return True

                except:
                    return False
            else:
                reactants_products = eq.split("->")
                reactants_products[1] += "+H"

                new_eq = "->".join(reactants_products)
                k, Ys = self.__lp_to_balance_reaction(new_eq)

                if len(k) == len(Ys) and 0 not in k.values():

                    new_reaction = reaction.copy()

                    new_reaction.add_metabolites({hydrogen: -1})

                    try:
                        stoich = self.__process_result_of_new_balanced_reaction(k, Ys, new_eq, new_reaction)

                        reaction.add_metabolites(stoich, combine=False)
                        return True

                    except:
                        return False

        elif len(reaction.compartments) == 1:

            reactants_products = eq.split("->")

            reactants = reactants_products[0].split("+")
            products = reactants_products[1].split("+")

            if "H" in reactants:
                reactants.remove("H")

                new_reactants = "+".join(reactants)
                new_eq = "->".join([new_reactants,reactants_products[1]])

                k, Ys = self.__lp_to_balance_reaction(new_eq)

                if len(k) == len(Ys):

                    new_reaction = reaction.copy()

                    new_reaction.subtract_metabolites({hydrogen: -1})
                    try:
                        stoich = self.__process_result_of_new_balanced_reaction(k, Ys, new_eq, new_reaction)

                        reaction.subtract_metabolites({hydrogen: -1})
                        reaction.add_metabolites(stoich, combine=False)
                        return True

                    except:

                        return False

            else:
                products.remove("H")

                new_products = "+".join(products)
                new_eq = "->".join([reactants_products[0], new_products])

                k, Ys = self.__lp_to_balance_reaction(new_eq)

                if len(k) == len(Ys):
                    new_reaction = reaction.copy()

                    new_reaction.subtract_metabolites({hydrogen: 1})
                    try:
                        stoich = self.__process_result_of_new_balanced_reaction(k, Ys, new_eq, new_reaction)

                        reaction.subtract_metabolites({hydrogen: 1})
                        reaction.add_metabolites(stoich, combine=False)
                        return True
                    except:
                        return False


        return False


    def __process_result_of_new_balanced_reaction(self,k,Ys,eq,reaction):

        N = []
        for s in sorted(Ys):
            value = k[Ys[s]]
            N.append(value)

        g = N[0]

        for a1, a2 in zip(N[0::2], N[1::2]):
            g = math.gcd(g, a2)

        N = [i / g for i in N]

        pM = lambda c: str(c) if c != 1 else ''

        res = {}

        equation_reactants_products = eq.split('->')
        reactants_in_equation = equation_reactants_products[0]
        products_in_equation = equation_reactants_products[1]

        for reactant in reactants_in_equation.split('+'):

            for reactant2 in reaction.reactants:
                if reactant2.formula == reactant:
                    stoich = pM(N.pop(0))
                    if stoich:
                        coef = int(stoich)
                    else:
                        coef = 1

                    res[reactant2] = coef * -1
                    break

        for product in products_in_equation.split('+'):

            for product2 in reaction.products:
                if product2.formula == product:
                    stoich = pM(N.pop(0))
                    if stoich:
                        coef = int(stoich)
                    else:
                        coef = 1

                    res[product2] = coef
                    break

        return res


    def __get_reaction_equation(self,reaction):
        reactants = reaction.reactants
        products = reaction.products
        eq = ""

        i=0
        while i<len(reactants)-1:

            reactant_formula = reactants[i].formula.replace("+","").replace("-","")
            eq+=reactant_formula+"+"

            i+=1


        eq += reactants[i].formula + "->"

        j=0
        while j < len(products) - 1:
            product_formula = products[j].formula.replace("+","").replace("-","")
            eq += product_formula + "+"

            j += 1

        eq += products[j].formula
        return eq


    def __balance_reaction(self, reaction, diff):
        """
        This method tries to balance a given reaction by using hydrogens

        :param Model.reaction reaction: reaction to balance.
        :param int diff: hydrogen difference between reactants and products

        :returns boolean: whether a reaction is balanced or not
        """

        compartment = list(reaction.compartments)
        if len(compartment) == 1:
            stoich_add = {}
            stoich_subtr = {}
            hydrogen = self.__hydrogens_in_model.get(compartment[0])
            new_reaction = reaction.copy()
            if diff < 0:

                isHydrogenInReaction = self.__check_if_compound_in_reaction(reaction,hydrogen)
                if isHydrogenInReaction:
                    stoich_subtr[hydrogen] = diff
                    new_reaction.subtract_metabolites(stoich_subtr)
                else:
                    stoich_add[hydrogen] = -diff
                    new_reaction.add_metabolites(stoich_add)

                self.__changed_reaction = new_reaction
                return True
            else:
                isHydrogenInReaction = self.__check_if_compound_in_reaction(reaction, hydrogen)
                if isHydrogenInReaction:
                    stoich_subtr[hydrogen] = -diff
                    new_reaction.subtract_metabolites(stoich_subtr)
                else:
                    stoich_add[hydrogen] = diff
                    new_reaction.add_metabolites(stoich_add)

                self.__changed_reaction = new_reaction
                return True


        return False



    def __check_if_compound_in_reaction(self,reaction,compound):
        for metabolite_in_reaction in reaction.metabolites:
            if metabolite_in_reaction.id == compound.id:
                return True
        return False

    def get_changed_reaction(self):
        return self.__changed_reaction

    def __get_hydrogen_from_model(self):
        """
        This method set searches for the hydrogen in each compartment of the model.
        The hydrogens are used to balance reactions.
        """

        self.__hydrogens_in_model = {}

        compartments = self.__model.compartments
        metabolites_in_model = self.__model.metabolites

        i = 0
        while i<len(metabolites_in_model):

            metabolite = metabolites_in_model[i]
            if metabolite.formula == "H" and metabolite.compartment in compartments:
                self.__hydrogens_in_model[metabolite.compartment] = metabolite

            i+=1

    def check_balanced_reaction(self, reaction):

        """
        This method checks if a given reaction is balanced or not, if not it tries to balance it.

        :param Model.reaction reaction: reaction to check whether is balanced.

        :returns boolean: whether a reaction is balanced or not
        """


        isBalanced = self.__check_if_reaction_is_balanced(reaction)

        if not isBalanced:

            return self.balance_reaction(reaction)

        else:
            return True


    def balance_and_clean_unbalanced_reactions(self):

        for reaction in self.__model.reactions:

            balanced = self.check_balanced_reaction(reaction)

            go = (not balanced and reaction not in self.__model.exchanges and
                  reaction not in self.__model.demands and reaction.annotation.get("sbo") != "SBO:0000629")
            if go:
                # self.__universal_model.add_reactions([reaction.copy()])
                self.__model.remove_reactions([reaction])


    def __check_if_reaction_is_balanced(self,reaction):
        (reactants,products) = self.__extract_elements_of_reactions_compounds(reaction)
        for element in reactants:
            if element in products.keys():
                diff = reactants[element] - products[element]
                if diff != 0:
                    return False
            else:
                return False

        for element in products:
            if element not in reactants.keys():
                return False

        return True


    def __extract_elements_of_reactions_compounds(self, reaction):
        """
       This method extracts the atomic elements of each compound in reaction

       :param Model.reaction reaction: reaction to extract atomic elements.

       :returns tuple: (reactants atomic elements, products atomic elements)
       """

        reactants_elements = {}
        products_elements = {}

        stoich = reaction.metabolites

        for metabolite in stoich.keys():
            formula = metabolite.formula
            coef = stoich[metabolite]
            match = re.findall("[A-Z][a-z]*[0-9]*", formula)
            for m in match:
                element = re.findall("[A-Z][a-z]*", m)[0]
                stoi = re.findall("[0-9]*", m)
                while "" in stoi:
                    stoi.remove("")

                if stoi:
                    if coef < 0:
                        reactants_elements[element[0]] = int(stoi[0]) * coef * -1 + reactants_elements.get(element[0], 0)
                    else:
                        products_elements[element[0]] = int(stoi[0]) * coef + products_elements.get(element[0], 0)
                else:
                    if coef < 0:
                        reactants_elements[element[0]] = coef * -1 + reactants_elements.get(element[0], 0)
                    else:
                        products_elements[element[0]] = coef + products_elements.get(element[0], 0)

        return (reactants_elements,products_elements)


    def check_compounds_representation_and_balance_reactions(self, reactions:list):

        res = []
        self.__get_hydrogen_from_model()

        compounds_ontology = CompoundsDBAccessor()

        to_remove = []
        for reaction in reactions:

            reaction_ontology_reactants = []
            reaction_ontology_products = []

            for compound in reaction.reactants:
                aliases = AnnotationUtils.get_annotation_from_cobra_annotation(compound)
                found = False
                for key in aliases:
                    for alias in aliases[key]:
                        new_key = self.compoundsAnnotationConfigs.get(key)

                        if key == "BOIMMG":
                            id = self.get_boimmg_id_from_annotation(aliases[key])
                            container = compounds_ontology.get_node_by_ont_id(id)

                            reaction_ontology_reactants.append(container)
                            found = True
                            break

                        elif new_key in self.__mapper.compounds_aliases_indexation.keys():
                            entities = self.__mapper.compounds_aliases_indexation.get(new_key).get(alias)
                            if entities:
                                for entity in entities:
                                    entity_id = entity.id

                                    if entity_id in self.__mapper.boimmg_db_model_map:

                                        ont_id = self.__mapper.boimmg_db_model_map.get(entity_id)
                                        entity_container = compounds_ontology.get_node_by_ont_id(ont_id)
                                        reaction_ontology_reactants.append(entity_container)
                                        found = True
                                        break
                    if found:
                        break


            for compound in reaction.products:

                found = False
                aliases = AnnotationUtils.get_annotation_from_cobra_annotation(compound)
                for key in aliases:
                    new_key = self.compoundsAnnotationConfigs.get(key)
                    for alias in aliases[key]:

                        if key == "BOIMMG":

                            id = self.get_boimmg_id_from_annotation(aliases[key])

                            container = compounds_ontology.get_node_by_ont_id(id)

                            reaction_ontology_products.append(container)
                            found = True
                            break


                        elif new_key in self.__mapper.compounds_aliases_indexation.keys():

                            entities = self.__mapper.compounds_aliases_indexation.get(new_key).get(alias)
                            if entities:
                                for entity in entities:
                                    entity_id = entity.id

                                    if entity_id in self.__mapper.boimmg_db_model_map:

                                        ont_id = self.__mapper.boimmg_db_model_map.get(entity_id)
                                        entity_container = compounds_ontology.get_node_by_ont_id(ont_id)
                                        reaction_ontology_products.append(entity_container)
                                        found = True
                                        break

                    if found:
                        break

            # self.__check_generic_complete_relationship(reaction_ontology_reactants,reaction_ontology_products,reaction)

            # if reaction_ontology_reactants and reaction_ontology_products:
            balanced = self.check_balanced_reaction(reaction)

            if not balanced:
                to_remove.append(reaction)

            else:
                res.append(reaction)

                # if generalization:
                #     self.__check_if_generic_complete_relationship()

        # print("reactions to remove: ")
        # print(to_remove)
        self.__model.remove_reactions(to_remove)
        return res

    def balance_reactions(self,reactions):
        to_remove = []
        res = []
        for reaction in reactions:
            balanced = self.check_balanced_reaction(reaction)

            if not balanced:
                to_remove.append(reaction)

            else:
                res.append(reaction)
        self.__model.remove_reactions(to_remove)
        return res

    def get_boimmg_id_from_annotation(self,boimmg_id):

        if isinstance(boimmg_id,list):
            boimmg_id = boimmg_id[0]

        boimmg_construction = self.compoundsAnnotationConfigs.get("BOIMMG_ID_CONSTRUCTION")

        id = int(boimmg_id.replace(boimmg_construction,""))

        return id

    def __check_generic_complete_relationship(self,reactants,products,reaction):

        generic_reactants = []
        complete_reactants = []
        generic_products = []
        complete_products = []

        compounds_ontology = CompoundsDBAccessor()

        for reactant in reactants:
            if reactant.generic:
                generic_reactants.append(reactant)
            else:
                complete_reactants.append(reactant)

        for product in products:
            if product.generic:
                generic_products.append(product)
            else:
                complete_products.append(product)

        if len(complete_reactants) == len(generic_products) or len(generic_reactants) == len(complete_products):
            return True

        elif not generic_reactants and not generic_products:
            reactants_components = []
            products_components = []

            reactants_components_dict = {}
            products_components_dict = {}


            for complete_reactant in complete_reactants:
                reactant_components = compounds_ontology.get_predecessors_by_ont_id_rel_type(complete_reactant.id,"component_of")
                reactants_components.append(reactant_components)
                reactants_components_dict[complete_reactant.id] = reactant_components


            for complete_product in complete_products:
                product_components = compounds_ontology.get_predecessors_by_ont_id_rel_type(complete_product.id,"component_of")
                products_components.append(product_components)
                products_components_dict[complete_product.id] = products_components

            reactants_components = sorted(reactants_components)
            products_components = sorted(products_components)

            if reactants_components!=products_components:
                deleted = self.__delete_components_from_reaction(reactants_components,
                                                       products_components,reactants_components_dict,
                                                       products_components_dict,reaction)

                if deleted:
                    return True

                else:
                    return False

        return True

    def __delete_components_from_reaction(self,reactants_components,
                                                products_components,reactants_components_dict,
                                                products_components_dict,reaction):

        to_delete = []

        if len(reactants_components) < len(products_components):
            reaction_products = reaction.products
            for product in reaction_products:
                product_components = products_components_dict[product.id]

                componentInReactants = []
                for component in product_components:
                    componentInReactants.append(component in reactants_components)

                if not any(componentInReactants):
                    for component in product_components:
                        products_components.remove(component)

                    to_delete.append(product)
                else:
                    # if it is necessary to delete a required component
                    return False

            if reactants_components != products_components:
                return False

            for metabolite in to_delete:
                reaction.subtract_metabolites({metabolite:1})

        else:
            reaction_reactants = reaction.reactants
            for reactant in reaction_reactants:
                reactant_components = reactants_components_dict[reactant.id]

                componentInProducts = []
                for component in reactant_components:
                    componentInProducts.append(component in reactants_components)

                if not any(componentInProducts):
                    for component in reactant_components:
                        reactants_components.remove(component)

                    to_delete.append(reactant)
                else:
                    # if it is necessary to delete a required component
                    return False

            if reactants_components != products_components:
                return False

            for metabolite in to_delete:
                reaction.subtract_metabolites({metabolite: 1})

        return True


    def set_model_mapper(self, mapper):
        self.__mapper = mapper



from enum import Enum


class BOIMMGProperties(Enum):
     FORMULA = "formula"
     NAME = "name"
     MASS = "mass"
     CHARGE = "charge"
     GENERIC = "generic"
     SMILES = "smiles"
     INCHI = "inchi"
     INCHIKEY = "inchikey"



from rdkit.Chem import AllChem


def matchers_functions(f):
    def wraps(*args, **kwargs):
        return f(*args, **kwargs)

    return wraps


class SidechainChecker:


    def __call__(self, core_smart, mol, side_chain_atoms_rules={}, molecule_atoms_rules={}):

        atoms_checked = True
        molecule_checked = True


        # core_smart = MolFromSmarts(core_smart)
        sidechain_compound = AllChem.DeleteSubstructs(mol, core_smart)

        if side_chain_atoms_rules:
            idx_lst = mol.GetSubstructMatches(core_smart)

            if not idx_lst:
                return False

            tam = mol.GetNumAtoms()
            substructure_idx = []
            for idx in idx_lst:
                for i in range(0, tam):
                    if i not in idx:
                        substructure_idx.append(i)

            atoms_checked = self.__check_sidechain_atoms__(substructure_idx, mol, side_chain_atoms_rules)

        if molecule_atoms_rules:
            molecule_checked = self.__check_molecule_atoms__(molecule_atoms_rules,sidechain_compound)

        if molecule_checked and atoms_checked:
            return True
        else:
            return False

    def __check_sidechain_atoms__(self, vect, mol, matchers_atoms):
        # loop over the atoms we care about:
        boolean_list = []
        for idx in vect:
            atom = mol.GetAtomWithIdx(idx)
            # print(atom.GetAtomicNum())
            for matcher in matchers_atoms:
                if callable(matchers_atoms[matcher]):
                    check = matchers_functions(matchers_atoms[matcher])(atom)


                    boolean_list.append(check)

        if all(boolean_list):
            return True

        else:
            return False

    def __check_molecule_atoms__(self, matchers_molecules, sidechain):
        boolean_list = []
        for matcher in matchers_molecules:
            if callable(matchers_molecules[matcher]):
                check = matchers_functions(matchers_molecules[matcher])(sidechain)
                boolean_list.append(check)

        if all(boolean_list):
            return True
        else:
            return False


from boimmgpy.kegg.kegg_entity import KeggEntity
import re


class KeggCompound(KeggEntity):

    def __init__(self,entry):
        super().__init__(entry)
        self.smiles=""
        self.inchikey=""
        self.formula = ""
        self.__get_all_information()


    def __get_all_information(self):
        info = self.get_raw_data()

        for text in info.split("\n"):
            new_line = re.sub('\s+', "\t", text)

            if re.search("\ANAME", new_line):
                feature = new_line.split("\t")
                self.name = feature[1]

            elif re.search("\AFORMULA", new_line):
                feature = new_line.split("\t")
                self.formula = feature[1]

    def get_name(self):
        return self.name

    def get_formula(self):
        return self.formula

    def get_kegg_id(self):
        return self.id


    def get_smiles(self):
        return self.smiles

    def get_inchikey(self):
        return self.inchikey







class ModelSeedReaction:

    def __init__(self,reaction_id, name, equation, direction, ec_numbers, deltag, pathways, compounds, aliases, stoichiometry):
        self.id = reaction_id
        self.name = name
        self.equation = equation
        self.set_direction(direction)
        self.ec_numbers = ec_numbers
        self.deltag = deltag
        self.pathways = pathways
        self.compounds = compounds
        self.aliases = aliases
        self.__setReversibility()
        if "Name" in self.aliases.keys():
            del self.aliases["Name"]
        self.stoichiometry = stoichiometry

    def set_direction(self,direction):
        if direction == "=":
            self.direction = "<=>"
        elif direction == ">":
            self.direction = "=>"
        elif direction == "<":
            self.direction = "<="

    def getEcNumbers(self):
        return self.ec_numbers

    def getDirection(self):
        return self.direction

    def getEquation(self):
        return self.equation

    def getDbId(self):
        return self.id

    def getReactants(self):
        reactants = []
        for compound in self.stoichiometry:
            if self.stoichiometry[compound]<0:
                reactants.append(compound)

        return reactants

    def getProducts(self):
        products = []
        for compound in self.stoichiometry:
            if self.stoichiometry[compound] > 0:
                products.append(compound)

        return products

    def getCompounds(self):
        return self.compounds

    def getDeltaG(self):
        return self.deltag

    def getName(self):
        return self.name

    def getStoichiometry(self):
        return self.stoichiometry

    def getAliases(self):
        return self.aliases

    def getAliasesByDatabase(self,db):
        if db == "ModelSEED":
            return [self.getDbId()]

        if db in self.aliases.keys():
            return self.aliases.get(db)
        else:
            return None

    def __setReversibility(self):
        if self.direction == "=":
            self.reversibility = True

        else:
            self.reversibility= False

    def getReversibility(self):
        return self.reversibility